import os
from pathlib import Path

import click
import questionary as q

from vantage6.common import error, info, warning
from vantage6.common.client.node_client import NodeClient
from vantage6.common.context import AppContext
from vantage6.common.globals import (
    FILE_BASED_DATABASE_TYPES,
    HTTP_LOCALHOST,
    SERVICE_BASED_DATABASE_TYPES,
    InstanceType,
    NodePolicy,
    Ports,
    RequiredNodeEnvVars,
)

from vantage6.cli.common.new import new
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.utils import prompt_config_name


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Store this configuration in the system folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Store this configuration in the user folders. This is the default.",
)
def cli_node_new_configuration(
    name: str,
    system_folders: bool,
) -> None:
    """
    Create a new node configuration.

    Checks if the configuration already exists. If this is not the case
    a questionnaire is invoked to create a new configuration file.
    """
    name = prompt_config_name(name)
    dirs = AppContext.instance_folders(InstanceType.NODE, name, system_folders)
    default_data_dir = str(dirs["data"])
    default_log_dir = str(dirs["log"])
    new(
        config_producing_func=node_configuration_questionaire,
        config_producing_func_args=(default_data_dir, default_log_dir, name),
        name=name,
        system_folders=system_folders,
        type_=InstanceType.NODE,
    )


def node_configuration_questionaire(
    data_dir: str, log_dir: str, instance_name: str
) -> dict:
    """
    Questionary to generate a config file for the node instance.

    Parameters
    ----------
    data_dir : str
        Path to the data directory of the node instance.
    log_dir : str
        Path to the log directory of the node instance.
    instance_name : str
        Name of the node instance.

    Returns
    -------
    dict
        Dictionary with the new node configuration
    """
    config = q.unsafe_prompt(
        [
            {"type": "text", "name": "api_key", "message": "Enter given api-key:"},
            {
                "type": "text",
                "name": "hq_url",
                "message": "The base-URL of HQ:",
                "default": HTTP_LOCALHOST,
            },
        ]
    )
    # remove trailing slash from hq_url if entered by user
    config["hq_url"] = config["hq_url"].rstrip("/")

    # set default port to the https port if hq_url is https
    default_port = (
        str(Ports.HTTPS) if config["hq_url"].startswith("https") else str(Ports.DEV_HQ)
    )

    config = config | q.unsafe_prompt(
        [
            {
                "type": "text",
                "name": "port",
                "message": "Enter port to which HQ listens:",
                "default": default_port,
            },
            {
                "type": "text",
                "name": "api_path",
                "message": "Path of the api:",
                "default": "/api",
            },
            {
                "type": "text",
                "name": "task_dir",
                "message": "Task directory path:",
                "default": data_dir,
            },
        ]
    )
    config["databases"] = {"fileBased": [], "serviceBased": []}
    while q.confirm("Do you want to add a database?").unsafe_ask():
        db_label = q.select(
            "What type of database do you want to add?",
            choices=["File database", "Database reachable by URI"],
        ).unsafe_ask()

        if db_label == "File database":
            config["databases"]["fileBased"].append(_get_file_based_database_config())
        else:
            config["databases"]["serviceBased"].append(
                _get_service_based_database_config()
            )

    is_policies = q.confirm(
        "Do you want to limit the algorithms allowed to run on your node? This "
        "should always be done for production scenarios.",
        default=True,
    ).unsafe_ask()
    policies = {}
    if is_policies:
        info(
            "You can limit the algorithms that can run on your node in two ways: by "
            "allowing specific algorithms or by allowing all algorithms in a given "
            "algorithm store."
        )
        ask_single_algorithms = q.confirm(
            "Do you want to enter a list of allowed algorithms?"
        ).unsafe_ask()
        if ask_single_algorithms:
            policies[NodePolicy.ALLOWED_ALGORITHMS.value] = _get_allowed_algorithms()
        ask_algorithm_stores = q.confirm(
            "Do you want to allow algorithms from specific algorithm stores?"
        ).unsafe_ask()
        if ask_algorithm_stores:
            policies[NodePolicy.ALLOWED_ALGORITHM_STORES.value] = (
                _get_allowed_algorithm_stores()
            )
        if ask_single_algorithms and ask_algorithm_stores:
            require_both_whitelists = q.confirm(
                "Do you want to allow only algorithms that are both in the list of "
                "allowed algorithms *AND* are part of one of the allowed algorithm "
                "stores? If not, algorithms will be allowed if they are in either the "
                "list of allowed algorithms or one of the allowed algorithm stores.",
                default=True,
            ).unsafe_ask()
            policies["allow_either_whitelist_or_store"] = not require_both_whitelists
    if policies:
        config["policies"] = policies

    res = q.select(
        "Which level of logging would you like?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"],
    ).unsafe_ask()

    config["logging"] = {
        "level": res,
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "requests", "level": "warning"},
            {"name": "engineio.client", "level": "warning"},
        ],
        "volumeHostPath": log_dir,
    }

    # Check if we can login to HQ to retrieve collaboration settings
    client = NodeClient(
        instance_name,
        config["api_key"],
        hq_url=f"{config['hq_url']}:{config['port']}{config['api_path']}",
        auth_url=os.environ.get(RequiredNodeEnvVars.KEYCLOAK_URL.value),
    )
    try:
        client.authenticate()
    except Exception as e:
        error(f"Could not authenticate: {e}")
        error(
            "Please check your API key and if HQ and authentication service are online."
        )
        warning(
            "If you continue, you should provide your collaboration settings manually."
        )
        if q.confirm("Do you want to abort?", default=True).unsafe_ask():
            exit(0)

    if client.whoami is not None:
        encryption = client.is_encrypted_collaboration()
        # TODO when we build collaboration policies, update this to provide
        # the node admin with a list of all policies, and whether or not
        # to accept them
        q.confirm(
            f"Encryption is {'enabled' if encryption else 'disabled'}"
            f" for this collaboration. Accept?",
            default=True,
        ).unsafe_ask()
    else:
        encryption = q.confirm("Enable encryption?", default=True).unsafe_ask()

    private_key = (
        "" if not encryption else q.text("Path to private key file:").unsafe_ask()
    )

    config["encryption"] = {
        "enabled": encryption is True or encryption == "true",
        "private_key": private_key,
    }

    # pack the entire config in a dict with the 'node' key at top level
    return {"node": config}


def _get_file_based_database_config() -> dict:
    """
    Prompt the user for the file-based database configuration
    """
    db_label = _get_database_label()
    while True:
        db_path = q.text(
            "Path to the database file:",
        ).unsafe_ask()
        if Path(db_path).exists():
            break
        else:
            error("The path to the database file does not exist. Please try again.")
    db_path_resolved = Path(db_path).resolve()
    db_dir = db_path_resolved.parent
    db_filename = db_path_resolved.name
    db_type = q.select("Database type:", choices=FILE_BASED_DATABASE_TYPES).unsafe_ask()
    return {
        "name": db_label,
        "uri": db_path,
        "type": db_type,
        "volumePath": db_dir,
        "originalName": db_filename,
    }


def _get_service_based_database_config() -> dict:
    """
    Prompt the user for the service-based database configuration

    Returns
    -------
    dict
        Dictionary with the service-based database configuration
    """
    db_label = _get_database_label()
    db_uri = q.text(
        "Database URI:",
    ).unsafe_ask()
    db_type = q.select(
        "Database type:", choices=SERVICE_BASED_DATABASE_TYPES
    ).unsafe_ask()

    env_vars = {}
    info("You can add environment variables to the database configuration.")
    info("These variables will be available to the algorithms on your node.")
    info("Example: MY_POSTGRES_USER=vantage6, MY_POSTGRES_PASSWORD=vantage6")
    while q.confirm("Do you want to add an environment variable?").unsafe_ask():
        env_var_name = q.text(
            "Enter the name of the environment variable:"
        ).unsafe_ask()
        env_var_value = q.text(
            "Enter the value of the environment variable:"
        ).unsafe_ask()
        env_vars[env_var_name] = env_var_value

    return {
        "name": db_label,
        "uri": db_uri,
        "type": db_type,
        "env": env_vars,
    }


def _get_database_label() -> str:
    """
    Prompt the user for the label of the database
    """
    return q.text(
        "Enter unique label for the database:",
        default="default",
    ).unsafe_ask()


def _get_allowed_algorithms() -> list[str]:
    """
    Prompt the user for the allowed algorithms on their node

    Returns
    -------
    list[str]
        List of allowed algorithms or regular expressions to match them
    """
    info("Below you can add algorithms that are allowed to run on your node.")
    info(
        "You can use regular expressions to match multiple algorithms, or you can "
        "use strings to provide one algorithm at a time."
    )
    info("Examples:")
    info(r"^harbor2\.vantage6\.ai/demo/average$    Allow the demo average algorithm")
    info(
        r"^harbor2\.vantage6\.ai/algorithms/.*   Allow all algorithms from "
        "harbor2.vantage6.ai/algorithms"
    )
    info(
        r"^harbor2\.vantage6\.ai/demo/average@sha256:82becede...$    Allow a "
        "specific hash of average algorithm"
    )
    allowed_algorithms = []
    while True:
        algo = q.text(message="Enter your algorithm expression:").unsafe_ask()
        allowed_algorithms.append(algo)
        if not q.confirm(
            "Do you want to add another algorithm expression?", default=True
        ).unsafe_ask():
            break
    return allowed_algorithms


def _get_allowed_algorithm_stores() -> list[str]:
    """
    Prompt the user for the allowed algorithm stores on their node

    Returns
    -------
    list[str]
        List of allowed algorithm stores
    """
    info("Below you can add algorithm stores that are allowed to run on your node.")
    info(
        "You can use regular expressions to match multiple algorithm stores, or you can"
        " use strings to provide one algorithm store at a time."
    )
    info("Examples:")
    info(
        "https://store.uluru.vantage6.ai    Allow all algorithms from the "
        "community store"
    )
    info(
        r"^https://*\.vantage6\.ai$               Allow all algorithms from any "
        "store hosted on vantage6.ai"
    )
    allowed_algorithm_stores = []
    while True:
        store = q.text(message="Enter the URL of the algorithm store:").unsafe_ask()
        allowed_algorithm_stores.append(store)
        if not q.confirm(
            "Do you want to add another algorithm store?", default=True
        ).unsafe_ask():
            break
    return allowed_algorithm_stores



import click

from vantage6.common import info

from vantage6.cli.common.attach import attach_logs
from vantage6.cli.context import InstanceType
from vantage6.cli.globals import InfraComponentName
from vantage6.cli.k8s_config import select_k8s_config


@click.command()
@click.option("-n", "--name", default=None, help="Name of the configuration")
@click.option("--system", "system_folders", flag_value=True, help="Use system folders")
@click.option("--user", "system_folders", flag_value=False, help="Use user folders")
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--sandbox", "is_sandbox", flag_value=True, help="Attach to a sandbox environment"
)
def cli_node_attach(
    name: str | None,
    system_folders: bool,
    context: str,
    namespace: str,
    is_sandbox: bool,
) -> None:
    """
    Show the node logs in the current console.
    """
    info("Attaching to node logs...")
    k8s_config = select_k8s_config(context=context, namespace=namespace)
    attach_logs(
        name=name,
        instance_type=InstanceType.NODE,
        infra_component=InfraComponentName.NODE,
        system_folders=system_folders,
        k8s_config=k8s_config,
        is_sandbox=is_sandbox,
    )



from typing import Any

import questionary as q

from vantage6.common import info
from vantage6.common.context import AppContext
from vantage6.common.globals import (
    InstanceType,
)

from vantage6.cli.auth.new import generate_password
from vantage6.cli.configuration_create import add_common_backend_config


def hq_configuration_questionaire(
    instance_name: str, system_folders: bool
) -> dict[str, Any]:
    """
    Kubernetes-specific questionnaire to generate Helm values for HQ.

    Parameters
    ----------
    instance_name : str
        Name of the HQ instance.
    system_folders : bool
        Whether to use system folders or user folders.

    Returns
    -------
    dict[str, Any]
        dictionary with Helm values for the HQ configuration
    """
    dirs = AppContext.instance_folders(InstanceType.HQ, instance_name, system_folders)
    log_dir = dirs.get("log")

    # Initialize config with basic structure
    config = {
        "hq": {"keycloak": {}},
        "database": {},
        "ui": {},
        "rabbitmq": {},
        "prometheus": {},
    }

    config = add_common_backend_config(config, InstanceType.HQ, instance_name)

    config = _set_allowed_algorithm_stores(config)

    # set directory to store log files on host machine
    config["hq"]["logging"]["volumeHostPath"] = str(log_dir)

    # set strong password for RabbitMQ
    config["rabbitmq"]["password"] = generate_password()

    return config


def _set_allowed_algorithm_stores(config: dict) -> dict:
    """
    Prompt the user for the allowed algorithm stores on their HQ.

    Parameters
    ----------
    config : dict
        Configuration dictionary.

    Returns
    -------
    dict
        Configuration dictionary with allowed algorithm stores.
    """
    info(
        "As security setting in the UI, we only allow traffic from certain external "
        "domains. These are vantage6 HQ and relevant algorithm stores."
    )
    info("Limiting the allowed algorithm stores is recommended.")
    setup_allowed_stores = q.confirm(
        "In the UI, do you want to only allow algorithms from specific algorithm "
        "stores? ",
        default=True,
    ).unsafe_ask()
    if setup_allowed_stores:
        config["ui"]["allowedAlgorithmStores"] = q.text(
            "Enter the URLs of all the algorithm stores you want to allow "
            "(comma-separated). Use * to allow all stores (less secure):",
            default="https://store.uluru.vantage6.ai",
        ).unsafe_ask()
    return config



"""
Kubernetes utility functions for Vantage6 CLI.

This module provides utilities for handling Kubernetes client configuration,
especially for MicroK8s environments that may have SSL certificate issues.

The issue is that the python kubernetes client does not automatically use the
certificate from microk8s, so it is manually configured in this module. Note that this
is only an issue for the CLI and not for running nodes/HQs/stores/etc., because
those are started using the `kubectl` command, which includes the microk8scertificate
automatically.
"""

import base64
import ssl
from pathlib import Path
from urllib.parse import urlparse

from kubernetes import client, config
from kubernetes.config.config_exception import ConfigException

from vantage6.common import warning


def configure_kubernetes_client_for_microk8s(cfg: client.Configuration) -> None:
    """
    Configure the Kubernetes client to handle MicroK8s SSL certificate issues.

    This function detects if we're using MicroK8s and configures the client
    to handle SSL certificates appropriately for both development and production.

    Parameters
    ----------
    cfg : client.Configuration
        The Kubernetes client configuration to configure
    """
    # Check if we're using MicroK8s by looking at the current context
    if is_microk8s_context():
        _configure_microk8s_ssl(cfg)


def _configure_microk8s_ssl(cfg: client.Configuration) -> None:
    """
    Configure SSL settings for MicroK8s.

    This function handles MicroK8s SSL certificates in a secure way by:
    1. First trying to use the MicroK8s certificate directly
    2. If that fails, falling back to a more lenient but still secure approach
    """
    try:
        # Try to get the MicroK8s certificate and use it properly
        if cert_path := _get_microk8s_certificate_path():
            _configure_with_certificate(cert_path, cfg)
        else:
            warning(
                "MicroK8s certificate not found. You may run into errors when using "
                "the CLI."
            )

    except Exception as e:
        warning(f"Could not configure MicroK8s SSL settings: {e}")
        warning("You may run into errors when using the CLI.")


def _get_microk8s_certificate_path() -> Path | None:
    """
    Get the path to the MicroK8s certificate.

    Returns
    -------
    Path | None
        Path to the MicroK8s certificate if found, None otherwise
    """
    # Common MicroK8s certificate locations
    possible_paths = [
        Path.home() / ".kube" / "microk8s.crt",
        Path("/var/snap/microk8s/current/certs/ca.crt"),
        Path("/var/snap/microk8s/current/certs/server.crt"),
    ]

    for cert_path in possible_paths:
        if cert_path.exists():
            return cert_path

    return None


def _configure_with_certificate(cert_path: Path, cfg: client.Configuration) -> None:
    """
    Configure the Kubernetes client to use a specific certificate.

    Parameters
    ----------
    cert_path : Path
        Path to the certificate file
    """
    try:
        # Validate the certificate before using it
        if not _validate_certificate(cert_path):
            warning(
                f"Certificate {cert_path} appears to be invalid. You may run into "
                "errors when using the CLI."
            )
            return

        cfg.verify_ssl = True
        cfg.ssl_ca_cert = str(cert_path)

        # Apply the configuration to the default client
        client.Configuration.set_default(cfg)

    except Exception as e:
        warning(f"Failed to configure with certificate {cert_path}: {e}")
        warning("You may run into errors when using the CLI.")


def _validate_certificate(cert_path: Path) -> bool:
    """
    Validate that a certificate file is readable and appears to be a valid certificate.

    Parameters
    ----------
    cert_path : Path
        Path to the certificate file

    Returns
    -------
    bool
        True if the certificate appears valid, False otherwise
    """
    try:
        # Check if the file exists and is readable
        if not cert_path.exists() or not cert_path.is_file():
            return False

        # Try to read the certificate content
        with open(cert_path, "rb") as f:
            cert_data = f.read()

        # Basic validation: check if it looks like a PEM certificate
        if b"-----BEGIN CERTIFICATE-----" not in cert_data:
            return False

        # Try to parse the certificate with Python's ssl module
        if _validate_certificate_with_ssl(cert_data):
            return True

        # If we can't parse it with ssl module, at least check it has the right
        # structure
        return (
            b"-----BEGIN CERTIFICATE-----" in cert_data
            and b"-----END CERTIFICATE-----" in cert_data
        )

    except Exception as e:
        warning(f"Certificate validation failed: {e}")
        return False


def _validate_certificate_with_ssl(cert_data: bytes) -> bool:
    """
    Validate a certificate with the ssl module.

    Parameters
    ----------
    cert_data : bytes
        The certificate data

    Returns
    -------
    bool
        True if the certificate appears valid, False otherwise
    """
    try:
        # Extract the base64 part between BEGIN and END
        start_text = b"-----BEGIN CERTIFICATE-----"
        len_start_text = len(start_text)
        start = cert_data.find(start_text)
        end = cert_data.find(b"-----END CERTIFICATE-----")
        if start != -1 and end != -1:
            cert_b64 = (
                cert_data[start + len_start_text : end]
                .replace(b"\n", b"")
                .replace(b"\r", b"")
            )
            cert_der = base64.b64decode(cert_b64)
            ssl.DER_cert_to_PEM_cert(cert_der)
            return True
    except Exception:
        pass
    return False


def load_kubernetes_config_with_fallback() -> bool:
    """
    Load Kubernetes configuration with fallback for development environments.

    This function tries to load the Kubernetes configuration and handles
    common issues like SSL certificate problems in development environments.

    Returns
    -------
    bool
        True if configuration was loaded successfully, False otherwise
    """
    # Try to load in-cluster config first (for when running inside Kubernetes)
    try:
        config.load_incluster_config()
        return True
    except ConfigException:
        pass

    # Fallback to kubeconfig
    try:
        # Load kubeconfig into default config, then adjust CA if MicroK8s
        config.load_kube_config()
        cfg = client.Configuration.get_default_copy()
        configure_kubernetes_client_for_microk8s(cfg)
        client.Configuration.set_default(cfg)

        return True
    except ConfigException as exc:
        warning(f"Failed to load Kubernetes configuration: {exc}")
        return False


def create_kubernetes_apis_with_ssl_handling() -> tuple[
    client.CoreV1Api, client.BatchV1Api
]:
    """
    Create Kubernetes API clients with SSL handling for development environments.

    Returns
    -------
    tuple[client.CoreV1Api, client.BatchV1Api]
        Tuple of CoreV1Api and BatchV1Api clients
    """
    # Load configuration with fallback handling
    if not load_kubernetes_config_with_fallback():
        raise RuntimeError("Failed to load Kubernetes configuration")

    # Create API clients
    core_api = client.CoreV1Api()
    batch_api = client.BatchV1Api()

    return core_api, batch_api


def get_core_api_with_ssl_handling() -> client.CoreV1Api:
    """
    Get the CoreV1Api client with SSL handling for development environments.
    """
    core_api, _ = create_kubernetes_apis_with_ssl_handling()
    return core_api


def is_microk8s_context() -> bool:
    """
    Check if the current Kubernetes context is MicroK8s.

    Returns
    -------
    bool
        True if using MicroK8s context, False otherwise
    """
    try:
        _, active_context = config.list_kube_config_contexts()
        current_context = active_context.get("name", "") if active_context else ""
        return "microk8s" in current_context.lower()
    except Exception:
        return False


def replace_localhost_for_k8s(url_or_uri: str, k8s_node: str | None = None) -> str:
    """
    Replace localhost in a URL or URI with the appropriate hostname for Kubernetes
    access.

    This function replaces 'localhost' with:
    - 'host.docker.internal' for Docker Desktop (when k8s_node == "docker-desktop")
    - '172.17.0.1' for Linux Kubernetes environments (MicroK8s, etc.)

    Parameters
    ----------
    url_or_uri : str
        The URL or URI that may contain localhost
    k8s_node : str | None
        The Kubernetes node name. If None, the URL/URI is returned unchanged.
        If "docker-desktop", localhost is replaced with "host.docker.internal".
        Otherwise, localhost is replaced with "172.17.0.1".

    Returns
    -------
    str
        The URL/URI with localhost replaced if applicable, otherwise unchanged
    """
    if k8s_node is None:
        return url_or_uri

    parsed = urlparse(url_or_uri)
    if parsed.hostname == "localhost":
        if k8s_node == "docker-desktop":
            return url_or_uri.replace("localhost", "host.docker.internal")
        else:
            return url_or_uri.replace("localhost", "172.17.0.1")

    return url_or_uri



from __future__ import annotations

from vantage6.common.context import AppContext
from vantage6.common.globals import InstanceType

from vantage6.cli import __version__
from vantage6.cli.configuration_manager import AuthConfigurationManager
from vantage6.cli.globals import DEFAULT_API_SERVICE_SYSTEM_FOLDERS as S_FOL


class AuthContext(AppContext):
    """
    Context class for the keycloak authentication service.

    Parameters
    ----------
    instance_name : str
        Name of the configuration instance, corresponds to the filename
        of the configuration file.
    system_folders : bool, optional
        System wide or user configuration, by default S_FOL
    is_sandbox : bool, optional
        Whether the configuration is a sandbox configuration, by default False
    """

    # The auth configuration manager is aware of the structure of the auth
    # configuration file and makes sure only valid configuration can be loaded.
    INST_CONFIG_MANAGER = AuthConfigurationManager

    def __init__(
        self,
        instance_name: str,
        system_folders: bool = S_FOL,
        is_sandbox: bool = False,
    ):
        super().__init__(
            InstanceType.AUTH,
            instance_name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
        )
        self.log.info("vantage6 version '%s'", __version__)

    @classmethod
    def from_external_config_file(
        cls, path: str, system_folders: bool = S_FOL
    ) -> AuthContext:
        """
        Create an auth context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL

        Returns
        -------
        AuthContext
            Auth context object
        """
        return super().from_external_config_file(
            path,
            InstanceType.AUTH,
            system_folders,
        )

    @classmethod
    def config_exists(
        cls,
        instance_name: str,
        system_folders: bool = S_FOL,
        is_sandbox: bool = False,
    ) -> bool:
        """
        Check if a configuration file exists.

        Parameters
        ----------
        instance_name : str
            Name of the configuration instance, corresponds to the filename
            of the configuration file.
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        is_sandbox : bool, optional
            Whether the configuration is a sandbox configuration, by default False

        Returns
        -------
        bool
            Whether the configuration file exists or not
        """
        return super().base_config_exists(
            InstanceType.AUTH,
            instance_name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
        )

    @classmethod
    def available_configurations(
        cls, system_folders: bool = S_FOL, is_sandbox: bool = False
    ) -> tuple[list, list]:
        """
        Find all available auth configurations in the default folders.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        is_sandbox : bool, optional
            Whether the configuration is a sandbox configuration, by default False

        Returns
        -------
        tuple[list, list]
            The first list contains validated configuration files, the second
            list contains invalid configuration files.
        """
        return super().available_configurations(
            InstanceType.AUTH, system_folders, is_sandbox
        )



from collections.abc import Callable

from colorama import Fore, Style

from vantage6.common import ensure_config_dir_writable, error, info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import get_main_cli_command_name
from vantage6.cli.configuration_create import make_configuration
from vantage6.cli.context import select_context_class
from vantage6.cli.utils import check_config_name_allowed, prompt_config_name


def new(
    config_producing_func: Callable,
    config_producing_func_args: tuple,
    name: str,
    system_folders: bool,
    type_: InstanceType,
    is_sandbox: bool = False,
    extra_config: dict | None = None,
    save_config_file: bool = True,
    extra_configs_to_render: dict[str, str] | None = None,
) -> dict | None:
    """
    Create a new configuration.

    Parameters
    ----------
    config_producing_func : Callable
        Function to generate the configuration
    config_producing_func_args : tuple
        Arguments to pass to the config producing function
    name : str
        Name of the configuration
    system_folders : bool
        Whether to store the configuration in the system folders
    type_ : InstanceType
        Type of the configuration (node, HQ, algorithm store, etc)
    is_sandbox : bool
        Whether to create a sandbox configuration or not
    extra_config: dict | None = None
        Extra configuration to add. Note that this may overwrite the configuration
        produced by the config producing function if the keys overlap.
    save_config_file: bool = True,
        Whether to store the configuration on disk or not.
    extra_configs_to_render: dict[str, str] | None = None
        Extra configurations to render. These will be added to the configuration file.

    Returns
    -------
    dict | None
        Dict with the configuration. None if the process is aborted for any reason.
    """
    name = prompt_config_name(name)

    # check if name is valid
    check_config_name_allowed(name)

    # check that this config does not exist
    ctx_class = select_context_class(type_)
    try:
        if ctx_class.config_exists(name, system_folders, is_sandbox):
            error(f"Configuration {Fore.RED}{name}{Style.RESET_ALL} already exists!")
            exit(1)
    except Exception as e:
        error(e)

        exit(1)

    command_name = get_main_cli_command_name(type_)

    # Check that we can write in this folder
    if not ensure_config_dir_writable(system_folders):
        error("Your user does not have write access to all folders. Exiting")
        info(
            f"Create a new {command_name} using '{Fore.GREEN}v6 {command_name} new "
            f"--user{Style.RESET_ALL}' instead!"
        )
        exit(1)

    # create config in ctx location
    try:
        config, cfg_file = make_configuration(
            config_producing_func=config_producing_func,
            config_producing_func_args=config_producing_func_args,
            type_=type_,
            instance_name=name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
            extra_config=extra_config,
            save_config_file=save_config_file,
            extra_configs_to_render=extra_configs_to_render,
        )
    except KeyboardInterrupt:
        error("Configuration creation aborted.")
        exit(1)
    if save_config_file:
        info(f"New configuration created: {Fore.GREEN}{cfg_file}{Style.RESET_ALL}")
        flag = "" if system_folders else "--user"
        info(
            f"You can start the {command_name} by running {Fore.GREEN}v6 {command_name}"
            f" start --name {name} {flag}{Style.RESET_ALL}"
        )
    else:
        info(
            "Configuration obtained successfully for "
            f"{Fore.GREEN}{name}{Style.RESET_ALL}."
        )

    return config



import click

from vantage6.common import info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.attach import attach_logs
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.start import (
    helm_install,
    prestart_checks,
)
from vantage6.cli.common.utils import create_directory_if_not_exists
from vantage6.cli.context.hq import HQContext
from vantage6.cli.globals import ChartName, InfraComponentName
from vantage6.cli.k8s_config import select_k8s_config


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--attach/--detach",
    default=False,
    help="Print HQ logs to the console after start",
)
@click.option("--local-chart-dir", default=None, help="Local chart directory to use")
@click.option("--chart-version", default=None, help="Chart version to use")
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(
    type_=InstanceType.HQ,
    include_name=True,
    include_system_folders=True,
    sandbox_param="sandbox",
)
def cli_hq_start(
    ctx: HQContext,
    name: str,
    system_folders: bool,
    context: str,
    namespace: str,
    attach: bool,
    local_chart_dir: str | None,
    chart_version: str | None,
) -> None:
    """
    Start an instance of vantage6 HQ.
    """
    info("Starting HQ...")
    prestart_checks(ctx, InstanceType.HQ, name, system_folders)

    k8s_config = select_k8s_config(context=context, namespace=namespace)

    create_directory_if_not_exists(ctx.log_dir)

    helm_install(
        release_name=ctx.helm_release_name,
        chart_name=ChartName.HQ,
        values_file=ctx.config_file,
        k8s_config=k8s_config,
        local_chart_dir=local_chart_dir,
        chart_version=chart_version,
    )

    if attach:
        attach_logs(
            name,
            instance_type=InstanceType.HQ,
            infra_component=InfraComponentName.HQ,
            system_folders=system_folders,
            k8s_config=k8s_config,
            is_sandbox=ctx.is_sandbox,
        )



import click

from vantage6.common import error
from vantage6.common.globals import InstanceType

from vantage6.cli.common.stop import execute_cli_stop
from vantage6.cli.common.utils import (
    find_running_service_names,
    get_config_name_from_helm_release_name,
    select_running_service,
)
from vantage6.cli.context import get_context
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import CLICommandName
from vantage6.cli.k8s_config import select_k8s_config


@click.command()
@click.option("-n", "--name", default=None, help="Name of the configuration.")
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
def cli_sandbox_stop(
    name: str | None,
    context: str | None,
    namespace: str | None,
) -> None:
    """
    Stop a sandbox environment.
    """
    k8s_config = select_k8s_config(context=context, namespace=namespace)

    running_services = find_running_service_names(
        instance_type=InstanceType.HUB,
        only_system_folders=False,
        only_user_folders=False,
        k8s_config=k8s_config,
    )

    if not running_services:
        error("No running sandbox services found.")
        return

    if not name:
        selected_service = select_running_service(running_services, InstanceType.HUB)
        name = get_config_name_from_helm_release_name(selected_service)
    else:
        ctx = get_context(InstanceType.HUB, name, False, is_sandbox=True)
        name = ctx.name

    # stop the sandbox nodes
    nodes_user_folder, _ = NodeContext.available_configurations(False, is_sandbox=True)
    for node in nodes_user_folder:
        if node.name.startswith(f"{name}-node-"):
            execute_cli_stop(
                command_name=CLICommandName.NODE,
                name=node.name,
                k8s_config=k8s_config,
                system_folders=False,
                is_sandbox=True,
            )

    execute_cli_stop(
        command_name=CLICommandName.HUB,
        name=name,
        k8s_config=k8s_config,
        system_folders=False,
        is_sandbox=True,
    )



import uuid
from typing import Any
from urllib.parse import urlparse

import questionary as q

from vantage6.common import error, info, warning
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import create_kubernetes_secret, generate_password
from vantage6.cli.configuration_create import (
    get_external_database_url,
)
from vantage6.cli.globals import APPNAME
from vantage6.cli.hub.utils.enum import AuthCredentials
from vantage6.cli.k8s_config import KubernetesConfig
from vantage6.cli.utils_kubernetes import (
    get_core_api_with_ssl_handling,
    replace_localhost_for_k8s,
)


def auth_configuration_questionaire(
    name: str,
    k8s_cfg: KubernetesConfig,
    credentials: dict[AuthCredentials, Any],
    include_global_auth_settings: bool = True,
) -> tuple[dict[str, Any]]:
    """
    Kubernetes-specific questionnaire to generate Helm values for the Keycloak helm
    chart.

    Parameters
    ----------
    name: str
        The name of the authentication service
    k8s_cfg: KubernetesConfig
        The Kubernetes configuration
    credentials: dict[AuthCredentials, Any]
        Dictionary with the credentials for the authentication service. This will be
        updated with the new credentials.
    include_global_auth_settings: bool
        Whether to include the global authentication settings. May be set to False
        if we are using the hub chart.

    Returns
    -------
    tuple[dict[str, Any]]
        Dictionary with the configuration for the authentication service
    """
    config = {"keycloak": {}, "database": {}}

    config["keycloak"]["production"] = True

    config = _add_external_database_config(config, k8s_node=k8s_cfg.k8s_node)

    config = _add_keycloak_admin_secret(config, name, k8s_cfg, credentials)

    if include_global_auth_settings:
        config["keycloak"].update(global_auth_settings_questionaire())

    # Add SMTP configuration if requested
    config = _add_smtp_config(config)

    return config


def global_auth_settings_questionaire() -> dict[str, Any]:
    """
    Questionaire to get the global authentication settings.

    These are the settings that are set in the global settings of the hub chart. They
    can also be used by the standalone auth chart, which can also use this function
    """
    return {
        "adminClientSecret": _get_admin_client_secret(),
        "adminPassword": _get_admin_password(),
    }


def _get_admin_client_secret() -> str:
    """
    Get the admin client secret.
    """
    provide_secret = q.confirm(
        "Do you want to choose a secret for the Keycloak backend client yourself? ",
        default=False,
    ).unsafe_ask()

    if provide_secret:
        admin_client_secret = q.password(
            "Enter Keycloak admin client secret:"
        ).unsafe_ask()
    else:
        admin_client_secret = generate_password(password_length=32)

    return admin_client_secret


def _get_admin_password() -> str:
    """
    Get the admin password.
    """
    return generate_password()


def _add_smtp_config(config: dict) -> dict:
    """
    Get SMTP server configuration from user.

    Returns
    -------
    dict
        Configuration dict with added SMTP configuration
    """
    configure_smtp = q.confirm(
        "Do you want to configure an SMTP server for email sending?",
        default=False,
    ).unsafe_ask()

    if not configure_smtp:
        return config

    smtp_config = {}

    smtp_config["host"] = q.text(
        "SMTP server hostname:",
        default="smtp.example.com",
    ).unsafe_ask()

    smtp_config["port"] = q.text(
        "SMTP server port (usually 587 for STARTTLS, 465 for SSL):",
        default="587",
    ).unsafe_ask()

    # SSL and STARTTLS are typically mutually exclusive
    # Port 465 usually uses SSL, port 587 usually uses STARTTLS
    if smtp_config["port"] == "587":
        encryption_type = "starttls"
    elif smtp_config["port"] == "465":
        encryption_type = "ssl"
    else:
        encryption_type = q.select(
            "SMTP encryption:",
            choices=[
                "starttls",
                "ssl",
                "none",
            ],
            default="starttls",
        ).unsafe_ask()

    if encryption_type == "starttls":
        smtp_config["ssl"] = "false"
        smtp_config["starttls"] = "true"
    elif encryption_type == "ssl":
        smtp_config["ssl"] = "true"
        smtp_config["starttls"] = "false"
    else:
        smtp_config["ssl"] = "false"
        smtp_config["starttls"] = "false"

    use_auth = q.confirm(
        "Does the SMTP server require authentication?",
        default=True,
    ).unsafe_ask()
    smtp_config["auth"] = "true" if use_auth else "false"

    if use_auth:
        # We currently only support basic (password-based) authentication. Bearer token
        # authentication is not supported - there is no need for it yet in our use case.
        smtp_config["authType"] = "basic"

        smtp_config["user"] = q.text(
            "SMTP username:",
        ).unsafe_ask()

        smtp_config["password"] = q.password(
            "SMTP password:",
        ).unsafe_ask()

    smtp_config["from"] = q.text(
        "Mail address used in the 'From' header:",
        default=smtp_config["user"] if smtp_config["user"] else "noreply@vantage6.ai",
    ).unsafe_ask()

    from_display_name = q.text(
        "Display name used in the 'From' header (optional, press Enter to skip):",
        default="",
    ).unsafe_ask()
    if from_display_name:
        smtp_config["fromDisplayName"] = from_display_name

    reply_to = q.text(
        "Reply-to email address (optional, press Enter to skip):",
        default="",
    ).unsafe_ask()
    if reply_to:
        smtp_config["replyTo"] = reply_to

    reply_to_display_name = q.text(
        "Reply-to display name (optional, press Enter to skip):",
        default="",
    ).unsafe_ask()
    if reply_to_display_name:
        smtp_config["replyToDisplayName"] = reply_to_display_name

    config["keycloak"]["smtpServer"] = smtp_config

    # if we add email server, we can also enable reset password and verify email
    config["keycloak"]["resetPasswordAllowed"] = True
    config["keycloak"]["verifyEmail"] = q.confirm(
        "Do you want to require users to verify their email address?",
        default=True,
    ).unsafe_ask()

    return config


def _add_external_database_config(config: dict, k8s_node: str | None = None) -> dict:
    """
    Add external database configuration for Keycloak.

    Keycloak uses separate fields (host, database, username, password) instead of a URI.
    This function prompts for a database URI and parses it to extract the required fields.

    Parameters
    ----------
    config : dict
        The configuration dictionary
    k8s_node : str | None
        The Kubernetes node name. Used to replace localhost with appropriate hostname.

    Returns
    -------
    dict
        The configuration dictionary with external database settings added
    """
    info("For production environments, it is recommended to use an external database.")
    info("Please provide the URI of the external database.")
    info("Example: postgresql://username:password@localhost:5432/keycloak")

    database_uri = get_external_database_url(InstanceType.AUTH)

    config["database"] = parse_database_uri_to_config(database_uri, k8s_node=k8s_node)

    return config


def parse_database_uri_to_config(
    database_uri: str, k8s_node: str | None = None
) -> dict:
    """
    Parse a database URI to a configuration dictionary.

    Parameters
    ----------
    database_uri : str
        The database URI to parse
    k8s_node : str | None
        The Kubernetes node name. If provided and hostname is 'localhost',
        it will be replaced with 'host.docker.internal' for Docker Desktop
        or '172.17.0.1' for Linux.

    Returns
    -------
    dict
        Dictionary with parsed database configuration
    """
    try:
        # Replace localhost in URI if needed
        database_uri = replace_localhost_for_k8s(database_uri, k8s_node)

        parsed = urlparse(database_uri)
        username = parsed.username
        password = parsed.password
        hostname = parsed.hostname
        port = parsed.port or 5432
        database_name = parsed.path.lstrip("/")

        for required_field in [
            ("username", username),
            ("password", password),
            ("hostname", hostname),
            ("database name", database_name),
        ]:
            if not required_field[1]:
                error(
                    f"No {required_field[0]} could be found in the database URI. "
                    f"Please provide a {required_field[0]}."
                )
                exit(1)
        if not port:
            warning(
                "No port could be found in the database URI. Using default port 5432."
            )

        return {
            "external": True,
            "host": hostname,
            "port": port,
            "name": database_name,
            "username": username,
            "password": password,
        }

    except Exception as e:
        error(f"Failed to parse database URI: {e}")
        error(
            "Please use format: postgresql://username:password@hostname:port/database"
        )
        exit(1)


def _add_keycloak_admin_secret(
    config: dict[str, Any],
    name: str,
    k8s_cfg: KubernetesConfig,
    credentials: dict[AuthCredentials, Any],
) -> dict[str, Any]:
    """
    Add the Keycloak admin secret to the config.

    Parameters
    ----------
    config: dict[str, Any]
        The configuration for the authentication service
    name: str
        The name of the authentication service
    k8s_cfg: KubernetesConfig
        The Kubernetes configuration
    credentials: dict[AuthCredentials, Any]
        Dictionary with the credentials for the authentication service. This will be
        updated with the new credentials.

    Returns
    -------
    dict[str, Any]
        The updated configuration for the authentication service
    """
    admin_user = q.text(
        "Keycloak admin username:",
        default="admin",
    ).unsafe_ask()

    provide_password = q.confirm(
        "Do you want to provide the Keycloak admin password yourself? Choosing 'no' "
        "will store a strong password in a Kubernetes secret and print it once.",
        default=False,
    ).unsafe_ask()

    if provide_password:
        admin_password = q.password(
            "Enter Keycloak admin password (stored in a Kubernetes secret):"
        ).unsafe_ask()
    else:
        admin_password = generate_password()

    # Determine namespace/context and create a Kubernetes Secret with the credentials
    try:
        secret_name = (
            f"{APPNAME}-{name}-{InstanceType.AUTH.value}-kc-admin-user-"
            f"{uuid.uuid4().hex[:8]}"
        )
        create_kubernetes_secret(
            core_api=get_core_api_with_ssl_handling(),
            secret_name=secret_name,
            namespace=k8s_cfg.namespace,
            secret_data={"username": admin_user, "password": admin_password},
        )
    except Exception as exc:
        error(f"Failed to create Keycloak admin secret: {exc}")
        exit(1)

    credentials[AuthCredentials.KEYCLOAK_ADMIN_USER] = admin_user
    credentials[AuthCredentials.KEYCLOAK_ADMIN_PASSWORD] = admin_password

    config["keycloak"]["adminUserSecret"] = secret_name

    return config



from dataclasses import dataclass
from importlib import resources as impresources
from pathlib import Path

import pandas as pd
from colorama import Fore, Style

from vantage6.common import error, info
from vantage6.common.globals import InstanceType, Ports

import vantage6.cli.sandbox.data as node_datafiles_dir
from vantage6.cli.common.new import new
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import (
    DefaultDatasets,
)
from vantage6.cli.k8s_config import KubernetesConfig
from vantage6.cli.sandbox.config.base import BaseSandboxConfigManager
from vantage6.cli.sandbox.populate.helpers.utils import (
    path_to_linux_style_str,
    replace_wsl_path,
    str_replace_wsl_path,
)


@dataclass
class NodeDataset:
    label: str
    path: Path


class NodeSandboxConfigManager(BaseSandboxConfigManager):
    """
    Class to store the node sandbox configurations.

    Parameters
    ----------
    hub_name : str
        Name of the HQ.
    api_keys : list[str]
        List of API keys.
    node_names : list[str]
        List of node names.
    hq_port : int
        Port of the HQ.
    node_image : str | None
        Image of the node.
    extra_node_config : Path | None
        Path to the extra node configuration file.
    extra_dataset : NodeDataset | None
        List of tuples with the label and path to the dataset file.
    k8s_config : KubernetesConfig
        Kubernetes configuration.
    custom_data_dir : Path | None
        Path to the custom data directory. Useful on WSL because of mount issues for
        default directories.
    with_prometheus : bool
        Whether Prometheus is enabled for the node.
    """

    def __init__(
        self,
        hub_name: str,
        api_keys: list[str],
        node_names: list[str],
        hq_port: int,
        node_image: str | None,
        extra_node_config: Path | None,
        extra_dataset: NodeDataset | None,
        k8s_config: KubernetesConfig,
        custom_data_dir: Path | None,
        with_prometheus: bool = False,
    ) -> None:
        super().__init__(hub_name, custom_data_dir)
        self.api_keys = api_keys
        self.node_names = node_names
        self.num_nodes = len(api_keys)
        self.hq_port = hq_port
        self.node_image = node_image
        self.extra_node_config = extra_node_config
        if extra_dataset:
            self.node_datasets = [extra_dataset]
        else:
            self.node_datasets = []
        self.k8s_config = k8s_config
        self.with_prometheus = with_prometheus

        self.node_configs = []
        self.node_config_names = []
        self.extra_config = None

    def generate_node_configs(self) -> None:
        """
        Generates ``num_nodes`` node configuration files.
        """
        node_data_files = []
        self.extra_config = self._read_extra_config_file(self.extra_node_config)

        data_directory = impresources.files(node_datafiles_dir)

        # Add default datasets to the list of dataset provided
        for default_dataset in DefaultDatasets:
            # note that the label is the name of the dataset with the underscores
            # replace with hyphens, to make it valid for k8s
            self.node_datasets.append(
                NodeDataset(
                    label=default_dataset.name.lower().replace("_", "-"),
                    path=data_directory / default_dataset.value,
                )
            )

        # Check for duplicate dataset labels
        seen_labels = set()
        duplicates = [
            label
            for label in [dataset.label for dataset in self.node_datasets]
            if (label in seen_labels or seen_labels.add(label))
        ]

        if len(duplicates) > 0:
            error(
                f"Duplicate dataset labels found: {duplicates}. "
                f"Please make sure all dataset labels are unique."
            )
            exit(1)

        # create the data files for the nodes and get the path and label for each
        # dataset
        for dataset in self.node_datasets:
            node_data_files.append(self._create_node_data_files(dataset))

        for idx, api_key in enumerate(self.api_keys):
            config = {
                "org_id": idx + 1,
                "api_key": api_key,
                "node_name": self.node_names[idx],
            }
            full_node_config = self._create_node_config_file(
                config,
                [files[idx] for files in node_data_files],
            )
            self.node_configs.append(full_node_config)
        info(
            f"Created {Fore.GREEN}{len(self.node_configs)}{Style.RESET_ALL} node "
            f"configuration(s), attaching them to {Fore.GREEN}{self.hub_name}"
            f"{Style.RESET_ALL}."
        )

    def _create_node_data_files(
        self, node_dataset: NodeDataset
    ) -> list[tuple[str, Path]]:
        """
        Create data files for nodes.

        Parameters
        ----------
        node_dataset : NodeDataset
            Tuple with the label and path to the dataset file.

        Returns
        -------
        list[tuple[str, Path]]
            List of tuples with the label and path to the dataset file.
        """
        info(
            f"Creating data files using dataset '{node_dataset.label}' for "
            f"{self.num_nodes} nodes"
        )
        data_files = []
        full_df = pd.read_csv(node_dataset.path)
        length_df = len(full_df)
        for i in range(self.num_nodes):
            node_name = f"{self.hub_name}_node_{i + 1}"
            path_to_dev_dir = self._create_and_get_data_dir(
                InstanceType.NODE, is_data_folder=False
            )

            # Split the data over the nodes
            start = i * length_df // self.num_nodes
            end = (i + 1) * length_df // self.num_nodes
            data = full_df[start:end]
            data_file = (
                replace_wsl_path(Path(path_to_dev_dir), to_mnt_wsl=True)
                / f"df_{node_dataset.label}_{node_name}.csv"
            )

            # write data to file
            data.to_csv(data_file, index=False)
            data_files.append(
                (node_dataset.label, str_replace_wsl_path(data_file, to_mnt_wsl=False))
            )
        return data_files

    def _create_node_config_file(
        self, config: dict, datasets: list[tuple[str, Path]]
    ) -> Path:
        """
        Create a node configuration file (YAML).

        Parameters
        ----------
        config : dict
            Configuration dictionary.
        datasets : list[tuple[str, Path]]
            List of tuples with the label and path to the dataset file.

        Returns
        -------
        Path
            Path to the node configuration file.
        """
        node_name = config["node_name"]
        config_name = f"{self.hub_name}-{node_name}"

        path_to_data_dir = self._create_and_get_data_dir(
            InstanceType.NODE, is_data_folder=True, node_name=node_name
        )
        log_dir = self._create_and_get_data_dir(
            InstanceType.NODE, is_log_dir=True, node_name=node_name
        )

        # delete old node config if it exists
        NodeContext.remove_config_file_if_exists(
            InstanceType.NODE, config_name, False, is_sandbox=True
        )

        # create new node config
        node_config = new(
            config_producing_func=self.__node_config_return_func,
            config_producing_func_args=(
                config,
                self.extra_config,
                datasets,
                path_to_data_dir,
                log_dir,
            ),
            name=config_name,
            system_folders=False,
            type_=InstanceType.NODE,
            is_sandbox=True,
        )
        self.node_config_names.append(config_name)

        return node_config

    def __node_config_return_func(
        self,
        node_specific_config: dict,
        extra_config: dict,
        datasets: list[tuple[str, Path]],
        path_to_data_dir: Path,
        log_dir: Path,
    ) -> dict:
        """
        Return a dict with node configuration values to be used in creating the
        config file.
        """
        config = {
            "node": {
                "proxyPort": 7676 + int(node_specific_config["org_id"]),
                "apiKey": node_specific_config["api_key"],
                "name": node_specific_config["node_name"],
                "logging": {
                    "level": "DEBUG",
                    "file": f"{node_specific_config['node_name']}.log",
                    "volumeHostPath": str(log_dir),
                },
                "keycloak": {
                    "url": (
                        f"http://vantage6-{self.hub_name}-user-hub-kc."
                        f"{self.k8s_config.namespace}.svc.cluster.local:"
                        f"{Ports.DEV_AUTH.value}"
                    ),
                },
                "persistence": {
                    "tasks": {
                        "hostPath": str(path_to_data_dir),
                        "size": "1Gi",
                    },
                    "database": {
                        "size": "1Gi",
                    },
                },
                "k8sNodeName": self.k8s_config.k8s_node,
                "databases": {
                    "fileBased": [
                        {
                            "name": dataset[0],
                            "uri": dataset[1],
                            "type": "csv",
                            # Call to path_to_linux_style_str is relevant for Windows
                            # paths in the sandbox environment.
                            "volumePath": path_to_linux_style_str(
                                Path(dataset[1]).parent
                            ),
                            "originalName": Path(dataset[1]).name,
                        }
                        for dataset in [datasets[0]]
                    ]
                },
                "hq": {
                    "url": (
                        f"http://vantage6-{self.hub_name}-user-hub-hq."
                        f"{self.k8s_config.namespace}.svc.cluster.local"
                    ),
                    "port": self.hq_port,
                },
            },
        }
        if self.node_image:
            config["node"]["image"] = self.node_image

        if self.with_prometheus:
            config["node"]["prometheus"] = {
                "enabled": True,
                "report_interval_seconds": 45,
            }

        # merge the extra config with the node config
        if extra_config is not None:
            config.update(extra_config)

        return config



from pathlib import Path

import click

from vantage6.common.globals import Ports

from vantage6.cli.sandbox.new import cli_new_sandbox
from vantage6.cli.sandbox.remove import cli_sandbox_remove
from vantage6.cli.sandbox.stop import cli_sandbox_stop
from vantage6.cli.test.feature_tester import cli_test_features
from vantage6.cli.utils import check_config_name_allowed, info, prompt_config_name


@click.command()
@click.option(
    "-n", "--name", default=None, type=str, help="Name for your development setup"
)
@click.option("--hq-image", type=str, default=None, help="HQ Docker image to use")
@click.option(
    "--store-image", type=str, default=None, help="Algorithm store Docker image to use"
)
@click.option("--node-image", type=str, default=None, help="Node Docker image to use")
@click.option(
    "--extra-hq-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional HQ configuration. This will be appended to the HQ "
    "configuration file",
)
@click.option(
    "--extra-store-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional algorithm store configuration. This will be "
    "appended to the algorithm store configuration file",
)
@click.option(
    "--extra-auth-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional auth configuration. This will be appended to the "
    "auth configuration file",
)
@click.option(
    "--extra-node-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional node configuration. This will be appended to the "
    "node configuration file",
)
@click.option(
    "--keep",
    type=bool,
    default=False,
    help="Keep the sandbox running after finishing the test",
)
@click.pass_context
def cli_test_integration(
    click_ctx: click.Context,
    name: str,
    hq_image: str,
    store_image: str,
    node_image: str,
    extra_hq_config: Path,
    extra_store_config: Path,
    extra_auth_config: Path,
    extra_node_config: Path,
    keep: bool = False,
) -> list[dict]:
    """
    Create sandbox network and run diagnostic checks on it.

    This is a full integration test of the vantage6 network. It will create
    a sandbox network with some nodes, and then run the v6-diagnostics algorithm
    to test all functionality.
    """
    # get name for the sandbox - if not given - and check if it is allowed
    name = prompt_config_name(name)
    check_config_name_allowed(name)

    # create sandbox and start it
    click_ctx.invoke(
        cli_new_sandbox,
        name=name,
        num_nodes=3,
        hq_image=hq_image,
        store_image=store_image,
        node_image=node_image,
        extra_hq_config=extra_hq_config,
        extra_store_config=extra_store_config,
        extra_auth_config=extra_auth_config,
        extra_node_config=extra_node_config,
    )

    # run the diagnostic tests
    diagnose_results = click_ctx.invoke(
        cli_test_features,
        hq_url=f"http://localhost:{Ports.SANDBOX_HQ}/hq",
        auth_url=f"http://localhost:{Ports.SANDBOX_AUTH}",
        auth_realm="vantage6",
        auth_client="public_client",
        collaboration=1,
        organizations=None,
        all_nodes=True,
        online_only=False,
        private_key=None,
        session_id=1,
        database_label="olympic-athletes",
    )

    # clean up the test resources
    if not keep:
        click_ctx.invoke(cli_sandbox_stop, name=name)
        click_ctx.invoke(cli_sandbox_remove, name=name)
    else:
        info(f"Keeping the sandbox {name}. You can stop it with `v6 sandbox stop`")

    return diagnose_results



from vantage6.common import error
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import (
    find_running_service_names,
    get_config_name_from_helm_release_name,
    select_running_service,
)
from vantage6.cli.context import get_context
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext
from vantage6.cli.context.hq import HQContext
from vantage6.cli.context.node import NodeContext
from vantage6.cli.k8s_config import select_k8s_config


def get_and_select_ctx(
    instance_type: InstanceType,
    name: str,
    system_folders: bool,
    context: str,
    namespace: str,
    is_sandbox: bool,
) -> HQContext | NodeContext | AlgorithmStoreContext:
    """
    Get and select the context for the given instance type.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to get the context for
    name : str
        The name of the instance
    system_folders : bool
        Whether to use system folders or not
    context : str
        The Kubernetes context to use
    namespace : str
        The Kubernetes namespace to use
    is_sandbox : bool
        Whether the configuration is a sandbox configuration

    Returns
    -------
    HQContext | NodeContext | AlgorithmStoreContext
        The context for the given instance type
    """
    k8s_config = select_k8s_config(
        context=context,
        namespace=namespace,
    )
    running_services = find_running_service_names(
        instance_type=instance_type,
        only_system_folders=False,
        only_user_folders=False,
        k8s_config=k8s_config,
    )

    if not running_services:
        error(f"No running {instance_type.value}s found.")
        exit(1)

    if not name:
        helm_name = select_running_service(running_services, instance_type)

        service_name = get_config_name_from_helm_release_name(
            helm_name, is_store=(instance_type == InstanceType.ALGORITHM_STORE)
        )
        ctx = get_context(
            instance_type, service_name, system_folders, is_sandbox=is_sandbox
        )

    else:
        ctx = get_context(instance_type, name, system_folders, is_sandbox=is_sandbox)
        helm_name = ctx.helm_release_name
        service_name = ctx.name

    if helm_name not in running_services:
        error(f"The {instance_type.value} {service_name} is not running.")
        exit(1)
    return ctx



"""
Utility functions for the CLI
"""

from __future__ import annotations

import os
import re
import subprocess
from pathlib import Path

import questionary as q

from vantage6.common import error, info, warning

MAX_LEN_NAME = 16


def _is_valid_k8s_dns_name(name: str) -> bool:
    """
    Validate a Kubernetes DNS-1035 label.

    Rules:
      - lower case alphanumeric or '-'
      - start with a letter
      - end with alphanumeric
      - length 1-63
    """
    if len(name) == 0 or len(name) > 63:
        return False
    return re.match(r"^[a-z]([-a-z0-9]*[a-z0-9])?$", name) is not None


def _check_k8s_dns_name(name: str) -> None:
    """
    Exit with error if the provided name is not a valid DNS-1035 label.
    """
    if not _is_valid_k8s_dns_name(name):
        error(
            f"Invalid name: '{name}'. Name must comply with the following rules: "
            "letters, numbers, or '-', start with a letter, end with letter or number, "
            "and be at most 63 characters."
        )
        exit(1)


def check_config_name_allowed(name: str) -> None:
    """
    Check if configuration name is allowed

    Parameters
    ----------
    name : str
        Name to be checked
    """
    if name.count(" ") > 0:
        name = name.replace(" ", "-")
        info(f"Replaced spaces from configuration name: {name}")
    if not re.match("^[a-zA-Z0-9_.-]+$", name):
        error(
            f"Name '{name}' is not allowed. Please use only the following "
            "characters: a-zA-Z0-9_.-"
        )
        exit(1)


def remove_file(file: str | Path, file_type: str) -> None:
    """
    Remove a file if it exists.

    Parameters
    ----------
    file : str
        absolute path to the file to be deleted
    file_type : str
        type of file, used for logging
    """
    if os.path.isfile(file):
        info(f"Removing {file_type} file: {file}")
        try:
            os.remove(file)
        except Exception as e:
            error(f"Could not delete file: {file}")
            error(e)
    else:
        warning(f"Could not remove {file_type} file: {file} does not exist")


def prompt_config_name(name: str | None) -> str:
    """
    Get a new configuration name from the user, or simply return the name if
    it is not None.

    Parameters
    ----------
    name : str
        Name to be checked

    Returns
    -------
    str
        The name of the configuration
    """
    if not name:
        try:
            name = q.text("Please enter a configuration-name:").unsafe_ask()
        except KeyboardInterrupt:
            error("Aborted by user!")
            exit(1)
        if not name:
            error("No configuration name provided!")
            exit(1)
        if name.count(" ") > 0:
            name = name.replace(" ", "-")
            info(f"Replaced spaces from configuration name: {name}")
    if len(name) > MAX_LEN_NAME:
        # Note that we set a limit of 16 chars because kubernetes has a limit of 63
        # chars for DNS labels. The remaining 47 chars may be appended and prepended by
        # vantage6 to identify a kubernetes resource.
        error(
            f"Configuration name '{name}' is too long! Maximum length is "
            f"{MAX_LEN_NAME} characters."
        )
        exit(1)
    _check_k8s_dns_name(name)
    return name


def switch_context_and_namespace(
    context: str | None = None, namespace: str | None = None
) -> None:
    # input validation
    validate_input_cmd_args(context, "context name", allow_none=True)
    validate_input_cmd_args(namespace, "namespace name", allow_none=True)

    try:
        if context:
            subprocess.run(
                ["kubectl", "config", "use-context", context],
                check=True,
                stdout=subprocess.DEVNULL,
            )
            info(f"Successfully set context to: {context}")

        if namespace:
            subprocess.run(
                [
                    "kubectl",
                    "config",
                    "set-context",
                    context or "--current",
                    f"--namespace={namespace}",
                ],
                check=True,
                stdout=subprocess.DEVNULL,
            )
            info(f"Successfully set namespace to: {namespace}")

    except subprocess.CalledProcessError as e:
        error(f"Failed to set Kubernetes context or namespace: {e}")


def validate_input_cmd_args(
    value: str | None, field_name: str, allow_none: bool = False
) -> None:
    """
    Validate input for subprocess commands.

    Exit with error if the input is invalid.

    Parameters
    ----------
    value : str | None
        The value to validate.
    field_name : str
        The name of the field being validated, used for error messages.
    allow_none : bool, optional
        Whether None is allowed as a valid value. Defaults to False.
    """
    if allow_none and value is None:
        return

    if not isinstance(value, str) or not re.match("^[a-zA-Z0-9_.-]+$", value):
        error(
            f"Invalid {field_name}: {value}. Use only alphanumeric characters, "
            "dashes, underscores, or dots."
        )
        exit(1)


def merge_nested_dicts(base: dict, extra: dict) -> dict:
    """
    Merge two nested dictionaries.

    If a key is not present in the base dictionary, it is added.
    If a key is present in the base dictionary and is a dictionary, it is merged.
    If a key is present in the base dictionary and is not a dictionary, it is replaced.

    Parameters
    ----------
    base: dict
        The base dictionary to merge into
    extra: dict
        The extra dictionary to merge from

    Returns
    -------
    dict
        The merged dictionary
    """
    for k, v in extra.items():
        if k in base and isinstance(base[k], dict) and isinstance(v, dict):
            merge_nested_dicts(base[k], v)  # merge in place
        else:
            base[k] = v
    return base



import click
from kubernetes import client as k8s_client
from kubernetes.client import ApiException

from vantage6.common import error, info, warning
from vantage6.common.globals import APPNAME, InstanceType

from vantage6.cli.common.stop import execute_stop, helm_uninstall
from vantage6.cli.common.utils import (
    extract_name_and_is_sandbox,
    get_config_name_from_helm_release_name,
)
from vantage6.cli.context import get_context
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import (
    DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL,
    InfraComponentName,
)
from vantage6.cli.k8s_config import KubernetesConfig
from vantage6.cli.node.common.task_cleanup import delete_run_related_pods
from vantage6.cli.utils_kubernetes import create_kubernetes_apis_with_ssl_handling


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders instead of user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in the user folders instead of "
    "system folders. This is the default.",
)
@click.option("--all", "all_nodes", flag_value=True, help="Stop all running nodes")
@click.option(
    "--sandbox", "is_sandbox", flag_value=True, help="Stop a sandbox environment"
)
def cli_node_stop(
    name: str,
    context: str,
    namespace: str,
    system_folders: bool,
    all_nodes: bool,
    is_sandbox: bool,
) -> None:
    """
    Stop one or all running nodes.
    """
    name, is_sandbox = extract_name_and_is_sandbox(name, is_sandbox)
    execute_stop(
        stop_function=_stop_node,
        stop_function_args={"system_folders": system_folders, "is_sandbox": is_sandbox},
        instance_type=InstanceType.NODE,
        infra_component=InfraComponentName.NODE,
        stop_all=all_nodes,
        to_stop=name,
        namespace=namespace,
        context=context,
        system_folders=system_folders,
        is_sandbox=is_sandbox,
    )


def _stop_node(
    node_helm_name: str,
    k8s_config: KubernetesConfig,
    system_folders: bool,
    is_sandbox: bool,
) -> None:
    """
    Stop a node

    Parameters
    ----------
    node_helm_name : str
        Name of the node to stop
    namespace : str
        Kubernetes namespace to use
    context : str
        Kubernetes context to use
    system_folders: bool
        Whether to use the system folders or not
    is_sandbox: bool
        Whether node is a sandbox node or not
    """
    helm_uninstall(release_name=node_helm_name, k8s_config=k8s_config)

    _stop_node_tasks(node_helm_name, system_folders, is_sandbox)

    info(f"Node {node_helm_name} stopped successfully.")


def _stop_node_tasks(
    node_helm_name: str, system_folders: bool, is_sandbox: bool
) -> None:
    """
    Stop the tasks of a node
    """
    node_name = get_config_name_from_helm_release_name(node_helm_name)
    node_ctx = get_context(
        InstanceType.NODE, node_name, system_folders, is_sandbox=is_sandbox
    )

    task_namespace = node_ctx.config.get("node", {}).get("taskNamespace")
    if not task_namespace:
        warning("Could not find node's task namespace. Node tasks will not be stopped.")
        return

    # detect tasks from the task namespace that this node is assigned to
    cleanup_task_jobs(task_namespace, node_ctx, all_nodes=False)


def cleanup_task_jobs(
    namespace: str, node_ctx: NodeContext | None = None, all_nodes: bool = False
) -> bool:
    """
    Cleanup Vantage6 task jobs in a given namespace.

    Parameters
    ----------
    namespace: str
        Namespace to cleanup
    node_ctx: NodeContext | None
        Node context to cleanup. If not given, all_nodes must be True.
    all_nodes: bool
        Cleanup all nodes. If not given, node_name must be given.

    Returns
    -------
    bool
        True if cleanup was successful, False otherwise
    """
    info(f"Cleaning up Vantage6 task jobs in namespace '{namespace}'")

    if not all_nodes and not node_ctx:
        error("Either all_nodes or node_ctx must be given to cleanup task jobs")
        return False

    # Load Kubernetes configuration with SSL handling
    try:
        core_api, batch_api = create_kubernetes_apis_with_ssl_handling()
    except RuntimeError as exc:
        error(f"Failed to load Kubernetes configuration: {exc}")
        return False

    jobs = _get_jobs(namespace, batch_api)

    deletions = 0
    for job in jobs:
        if not all_nodes and job.metadata.labels.get("node_id") != node_ctx.identifier:
            # if all_nodes is False, we should only delete jobs assigned to the current
            # node
            continue
        elif all_nodes and not _is_vantage6_task_job(job):
            # if all_nodes is True, we should only delete vantage6 task jobs, not other
            # jobs
            continue

        run_id = _get_job_run_id(job)
        if run_id is None:
            error(f"Job '{job.metadata.name}' has no run_id annotation, skipping...")
            continue

        # Use shared cleanup to delete job, pods and related secret
        job_name = job.metadata.name
        info(f"Deleting job '{job_name}' (run_id={run_id})")
        delete_run_related_pods(
            run_id=run_id,
            container_name=f"{APPNAME}-run-{run_id}",
            namespace=namespace,
            core_api=core_api,
            batch_api=batch_api,
        )
        deletions += 1

    if deletions == 0:
        info(f"No Vantage6 task jobs found to delete in namespace '{namespace}'")
    else:
        info(f"Deleted {deletions} Vantage6 task job(s) in namespace '{namespace}'")
    return True


def _is_vantage6_task_job(job: k8s_client.V1Job) -> bool:
    # Vantage6 task jobs can be identified by their name, which is of the form
    # "vantage6-run-<run_id>"
    return job.metadata.name.startswith(f"{APPNAME}-run-")


def _get_jobs(
    namespace: str, batch_api: k8s_client.BatchV1Api
) -> list[k8s_client.V1Job]:
    try:
        return batch_api.list_namespaced_job(namespace=namespace).items
    except ApiException as exc:
        error(f"Failed to list jobs in namespace {namespace}: {exc}")
        return []


def _get_job_run_id(job: k8s_client.V1Job) -> int | None:
    annotations = job.metadata.annotations or {}
    try:
        return int(annotations.get("run_id"))
    except ValueError:
        error(f"Job '{job.metadata.name}' has no run_id annotation, skipping")
        return None



from typing import Any
from urllib.parse import urlparse

import click
import questionary as q
import yaml
from colorama import Fore, Style

from vantage6.common import error, info
from vantage6.common.globals import (
    InstanceType,
    Ports,
)

from vantage6.cli.algostore.new import algo_store_configuration_questionaire
from vantage6.cli.auth.new import (
    auth_configuration_questionaire,
    global_auth_settings_questionaire,
)
from vantage6.cli.common.new import new
from vantage6.cli.context.hub import HubContext
from vantage6.cli.globals import DEFAULT_API_SERVICE_SYSTEM_FOLDERS
from vantage6.cli.hq.new import hq_configuration_questionaire
from vantage6.cli.hub.utils.enum import AuthCredentials
from vantage6.cli.k8s_config import (
    KubernetesConfig,
    get_k8s_node_names,
    select_k8s_config,
)
from vantage6.cli.utils import prompt_config_name


@click.command()
@click.option(
    "-n", "--name", default=None, help="name of the configuration you want to create."
)
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Use system folders instead of user folders. This is the default",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=DEFAULT_API_SERVICE_SYSTEM_FOLDERS,
    help="Use user folders instead of system folders",
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
def cli_hub_new(
    name: str,
    system_folders: bool,
    context: str | None,
    namespace: str | None,
) -> None:
    """
    Create production-ready configuration for a complete vantage6 hub.

    This will create production-ready configurations for the vantage6 hub's components,
    i.e. HQ, auth, algorithm store, ui, as well as related services such as RabbitMQ
    and Prometheus.
    """
    name = prompt_config_name(name)

    if HubContext.config_exists(name, system_folders):
        error(f"Configuration {Fore.RED}{name}{Style.RESET_ALL} already exists!")
        exit(1)

    k8s_cfg = select_k8s_config(context=context, namespace=namespace)

    # get basic general configuration (e.g. URLs for the services)
    info("Starting with the basic configuration...")
    base_config = _get_base_config()

    global_hub_config = _create_hub_config(name, base_config, k8s_cfg)

    # create authentication service configuration
    info("Now, let's setup the authentication service...")
    auth_name = f"{name}-auth"
    auth_credentials = {}
    extra_config = {
        "keycloak": {
            # Keycloak is exposed via a Gateway (TLS terminated at the Gateway)
            "behindGateway": {
                "enabled": True,
                "proxyHeaders": "xforwarded",
                "httpPort": 7680,
            },
            "hostname": base_config["auth_url"],
            "publicClient": {
                "redirectUris": [
                    f"{base_config['ui_url']}",
                ],
            },
        }
    }
    auth_config = new(
        config_producing_func=auth_configuration_questionaire,
        config_producing_func_args=(auth_name, k8s_cfg, auth_credentials, False),
        name=auth_name,
        system_folders=system_folders,
        type_=InstanceType.AUTH,
        extra_config=extra_config,
        save_config_file=False,
    )

    # create hq service configuration
    info("Now, let's setup the vantage6 HQ...")
    hq_name = name
    extra_config = {
        "hq": {
            "logging": {
                "level": base_config["log_level"],
            },
        },
    }
    if base_config["has_store"]:
        extra_config["hq"]["algorithm_stores"] = [
            {
                "name": base_config["store_name"],
                "url": base_config["store_url"],
                "api_path": "/store",
            }
        ]
        extra_config["ui"] = {
            "communityStoreApiPath": "/store",
        }
    hq_config = new(
        config_producing_func=hq_configuration_questionaire,
        config_producing_func_args=(hq_name, system_folders),
        name=hq_name,
        system_folders=system_folders,
        type_=InstanceType.HQ,
        extra_config=extra_config,
        save_config_file=False,
    )

    # create algorithm store service configuration
    store_config = None
    auth_config_dict = yaml.safe_load(auth_config)
    if base_config["has_store"]:
        info("Finally, let's setup the algorithm store...")
        store_name = f"{name}-store"
        extra_config = {
            "store": {
                "logging": {
                    "level": base_config["log_level"],
                },
                "api_path": "/store",
            }
        }
        if auth_config_dict.get("keycloak", {}).get("smtpServer") is not None:
            extra_config["store"]["smtpServer"] = {
                "host": auth_config_dict["keycloak"]["smtpServer"]["host"],
                "port": auth_config_dict["keycloak"]["smtpServer"]["port"],
                "from": auth_config_dict["keycloak"]["smtpServer"]["from"],
            }
            if auth_config_dict["keycloak"]["smtpServer"].get("user") is not None:
                extra_config["store"]["smtpServer"]["user"] = auth_config_dict[
                    "keycloak"
                ]["smtpServer"]["user"]
            if auth_config_dict["keycloak"]["smtpServer"].get("password") is not None:
                extra_config["store"]["smtpServer"]["password"] = auth_config_dict[
                    "keycloak"
                ]["smtpServer"]["password"]
            if auth_config_dict["keycloak"]["smtpServer"].get("replyTo") is not None:
                extra_config["store"]["smtpServer"]["replyTo"] = auth_config_dict[
                    "keycloak"
                ]["smtpServer"]["replyTo"]
            if auth_config_dict["keycloak"]["smtpServer"].get("starttls") is not None:
                extra_config["store"]["smtpServer"]["starttls"] = auth_config_dict[
                    "keycloak"
                ]["smtpServer"]["starttls"]
            if auth_config_dict["keycloak"]["smtpServer"].get("ssl") is not None:
                extra_config["store"]["smtpServer"]["ssl"] = auth_config_dict[
                    "keycloak"
                ]["smtpServer"]["ssl"]
        store_config = new(
            config_producing_func=algo_store_configuration_questionaire,
            config_producing_func_args=(store_name, system_folders),
            name=store_name,
            system_folders=system_folders,
            type_=InstanceType.ALGORITHM_STORE,
            extra_config=extra_config,
            save_config_file=False,
        )

    # create the final hub configuration file
    extra_configs_to_render = {
        "auth_config": auth_config,
        "hq_config": hq_config,
    }
    if store_config is not None:
        extra_configs_to_render["store_config"] = store_config
    new(
        config_producing_func=lambda: global_hub_config,
        config_producing_func_args=(),
        name=name,
        system_folders=system_folders,
        type_=InstanceType.HUB,
        save_config_file=True,
        extra_configs_to_render=extra_configs_to_render,
    )

    _print_credentials_one_time(auth_credentials, global_hub_config["keycloak"])


def _get_base_config() -> dict[str, Any]:
    """
    Get the base configuration for a vantage6 hub's components.
    """
    base_config = {}
    info(
        "Logs and prometheus data can currently only be stored on a specific k8s node."
    )
    k8s_node_names = get_k8s_node_names()
    no_local_storage = "Don't store logs and prometheus data"
    k8s_node_name = q.select(
        "In which k8s node do you want to store logs and prometheus data?",
        choices=[no_local_storage] + k8s_node_names,
        default=k8s_node_names[0],
    ).unsafe_ask()
    if k8s_node_name != no_local_storage:
        base_config["k8sNodeName"] = k8s_node_name
    info("We need you to provide a domain where you will deploy the hub services.")
    info("For instance, the domain 'example.com' will lead to the following URLs:")
    info("Authentication:  https://auth.example.com")
    info("HQ:              https://hq.example.com")
    info("UI:              https://portal.example.com")
    info("Algorithm store: https://store.example.com")
    url_domain = q.text(
        "On what domain will the services be reachable?",
        default="vantage6.ai",
    ).unsafe_ask()
    base_config["hq_url"] = f"https://hq.{url_domain}"
    base_config["auth_url"] = f"https://auth.{url_domain}"
    base_config["ui_url"] = f"https://portal.{url_domain}"
    base_config["has_store"] = q.confirm(
        "Do you want to use an algorithm store?",
        default=True,
    ).unsafe_ask()
    if base_config["has_store"]:
        base_config["store_url"] = f"https://store.{url_domain}"
        base_config["store_name"] = q.text(
            "What is the name of the algorithm store?",
            default="",
            validate=lambda x: x != "",
        ).unsafe_ask()
    base_config["log_level"] = q.select(
        "What is the log level for the algorithm store?",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    ).unsafe_ask()
    return base_config


def _create_hub_config(
    name: str, base_config: dict[str, Any], k8s_cfg: KubernetesConfig
) -> dict[str, Any]:
    """
    Create a hub configuration file (YAML) that aggregates the auth, HQ
    and algorithm store configurations, so the sandbox can be deployed
    via the single hub Helm chart.

    Parameters
    ----------
    name: str
        The name of the hub.
    base_config: dict[str, Any]
        The base configuration for the hub.
    k8s_cfg: KubernetesConfig
        The Kubernetes configuration to use.

    Returns
    -------
    dict[str, Any]
        The hub configuration.
    """

    def _hostname_from_url(url: str) -> str:
        parsed = urlparse(url)
        return parsed.hostname or ""

    urls = {
        "external": {
            "auth": base_config["auth_url"],
            "store": base_config.get("store_url", ""),
            # The UI is typically exposed on portal.<domain>
            "ui": base_config["ui_url"],
            "hq": base_config["hq_url"],
        },
        "internal": {
            "auth": (
                f"http://vantage6-{name}-user-hub-kc-service.{k8s_cfg.namespace}.svc"
                f".cluster.local:{Ports.DEV_AUTH.value}"
            ),
            "store": (
                f"http://vantage6-{name}-user-hub-store.{k8s_cfg.namespace}.svc"
                f".cluster.local:{Ports.DEV_ALGO_STORE.value}"
            ),
            "ui": (
                f"http://vantage6-{name}-user-hub-ui.{k8s_cfg.namespace}.svc"
                f".cluster.local:{Ports.DEV_UI.value}"
            ),
            "hq": (
                f"http://vantage6-{name}-user-hub-hq.{k8s_cfg.namespace}.svc"
                f".cluster.local:{Ports.DEV_HQ.value}"
            ),
        },
    }

    hub_gateway = {
        "enabled": True,
        "hosts": {
            "auth": _hostname_from_url(urls["external"]["auth"]),
            "hq": _hostname_from_url(urls["external"]["hq"]),
            "portal": _hostname_from_url(urls["external"]["ui"]),
            "store": _hostname_from_url(urls["external"]["store"]),
        },
        "tls": {
            "mode": "cert-manager",
            "existingSecrets": {
                "auth": "",
                "hq": "",
                "portal": "",
                "store": "",
            },
        },
        "certManager": {
            "enabled": True,
            "clusterIssuer": "letsencrypt-prod",
        },
    }

    keycloak = {
        "url": urls["external"]["auth"],
    } | global_auth_settings_questionaire()

    global_config = {
        "hubGateway": hub_gateway,
        "urls": urls,
        "keycloak": keycloak,
    }
    if base_config.get("k8sNodeName") is not None:
        global_config["k8sNodeName"] = base_config["k8sNodeName"]
    return global_config


def _print_credentials_one_time(
    credentials: dict[AuthCredentials, Any] | None, keycloak_config: dict
) -> None:
    """
    Print the used credentials one time.

    Parameters
    ----------
    credentials: dict[AuthCredentials, Any] | None
        Dictionary with the credentials for the authentication service.
    keycloak_config: dict
        Keycloak section of the auth configuration
    """
    if not credentials:
        return
    info("--------------------------------")
    info(
        "In setting up the service, you generated credentials that have been stored"
        " in Kubernetes secrets."
    )
    info(
        "Do NOT delete the Kubernetes secrets as long as you use this authentication "
        "service."
    )
    info("This is a one-time print of the credentials. They will not be printed again.")
    info("--------------------------------")
    for credential, value in credentials.items():
        info(f"{credential.description}: {value}")
    info("--------------------------------")
    info("You can login to vantage6 with the following credentials:")
    info(f"Username: {keycloak_config.get('adminUser', 'admin')}")
    info(f"Password: {keycloak_config['adminPassword']}")
    info("--------------------------------")



"""
Common functions that are used in node CLI commands
"""

import os

from colorama import Fore, Style

from vantage6.common import debug, error, info
from vantage6.common.globals import (
    HTTP_LOCALHOST,
    InstanceType,
    Ports,
    RequiredNodeEnvVars,
)

from vantage6.client import UserClient
from vantage6.client.utils import LogLevel

from vantage6.cli.configuration_create import select_configuration_questionnaire
from vantage6.cli.context.node import NodeContext


def _convert_k8s_url_to_localhost(url: str) -> str:
    """
    Convert a Kubernetes URL to a localhost URL.
    """
    if "svc.cluster.local" in url:
        port_and_api_path = url.split(":")[-1]
        return f"{HTTP_LOCALHOST}:{port_and_api_path}"
    return url


def create_client(ctx: NodeContext, use_sandbox_port: bool = False) -> UserClient:
    """
    Create a client instance.

    Parameters
    ----------
    ctx : NodeContext
        Context of the node loaded from the configuration file
    use_sandbox_port : bool
        Whether to use the sandbox port or not

    Returns
    -------
    UserClient
        vantage6 client
    """
    host = ctx.config["node"]["hq"]["url"]
    port = ctx.config["node"]["hq"]["port"]
    api_path = ctx.config["node"]["hq"]["path"]
    # if the hq is run locally in Docker, we need to use localhost here instead of
    # the host address of docker
    if host in ["http://host.docker.internal", "http://172.17.0.1"]:
        host = HTTP_LOCALHOST

    url = f"{host}:{port}{api_path}"

    auth_url = ctx.config.get("node", {}).get("keycloak", {}).get(
        "url", None
    ) or os.environ.get(RequiredNodeEnvVars.KEYCLOAK_URL.value)
    # append the port to the auth URL as it is not included in the config
    if use_sandbox_port:
        auth_url = f"{auth_url}:{Ports.SANDBOX_AUTH.value}"
    else:
        auth_url = f"{auth_url}:{Ports.DEV_AUTH.value}"

    # if the URL is a Kubernetes address, we need to use localhost because here
    # we are connecting from the CLI outside the cluster
    url = _convert_k8s_url_to_localhost(url)
    auth_url = _convert_k8s_url_to_localhost(auth_url)

    info(f"Connecting to HQ at '{url}' using auth URL '{auth_url}'")
    return UserClient(
        hq_url=url,
        auth_url=auth_url,
        log_level=LogLevel.WARN,
    )


def create_client_and_authenticate(
    ctx: NodeContext, use_sandbox_port: bool = False
) -> UserClient:
    """
    Generate a client and authenticate.

    Parameters
    ----------
    ctx : NodeContext
        Context of the node loaded from the configuration file
    use_sandbox_port : bool
        Whether to use the sandbox port or not

    Returns
    -------
    UserClient
        vantage6 client
    """
    client = create_client(ctx, use_sandbox_port)

    try:
        client.authenticate()
    except Exception as exc:
        error("Could not authenticate!")
        debug(str(exc))
        exit(1)

    return client


def select_node(name: str, system_folders: bool) -> str:
    """
    Let user select node through questionnaire if name is not given.

    Parameters
    ----------
    name : str
        Name of the node to select
    system_folders : bool
        Whether to use system folders or not

    Returns
    -------
    str
        Name of the configuration file
    """
    try:
        name = (
            name
            if name
            else select_configuration_questionnaire(InstanceType.NODE, system_folders)
        )
    except Exception:
        error("No configurations could be found!")
        exit()

    # raise error if config could not be found
    if not NodeContext.config_exists(name, system_folders):
        error(
            f"The configuration {Fore.RED}{name}{Style.RESET_ALL} could not be found."
        )
        exit(1)
    return name



from collections.abc import Callable
from pathlib import Path
from urllib.parse import urlparse

import questionary as q

from vantage6.common import error, info
from vantage6.common.context import AppContext
from vantage6.common.globals import (
    DEFAULT_API_PATH,
    InstanceType,
)

from vantage6.cli.configuration_manager import (
    AlgorithmStoreConfigurationManager,
    AuthConfigurationManager,
    HQConfigurationManager,
    HubConfigurationManager,
    NodeConfigurationManager,
)
from vantage6.cli.context import select_context_class
from vantage6.cli.utils import merge_nested_dicts


def add_common_backend_config(
    config: dict, instance_type: InstanceType, instance_name: str
) -> dict:
    """
    Part of the questionaire that is common to all backend types (HQ and algorithm
    store).

    Parameters
    ----------
    instance_type : InstanceType
        Type of backend instance.
    instance_name : str
        Name of the backend instance.

    Returns
    -------
    dict
        Dictionary with new (partial) backend configuration
    """
    service_name = "hq" if instance_type == InstanceType.HQ else "store"
    backend_config = config[service_name]

    backend_config["logging"] = {
        "file": f"{instance_name}-{service_name}.log",
        "use_console": True,
        "backup_count": 5,
        "max_size": 1024,
        "format": "%(asctime)s - %(name)-14s - %(levelname)-8s - %(message)s",
        "datefmt": "%Y-%m-%d %H:%M:%S",
        "loggers": [
            {"name": "urllib3", "level": "warning"},
            {"name": "socketIO-client", "level": "warning"},
            {"name": "socketio.server", "level": "warning"},
            {"name": "engineio.server", "level": "warning"},
            {"name": "sqlalchemy.engine", "level": "warning"},
        ],
    }

    backend_config["api_path"] = DEFAULT_API_PATH

    config["database"]["external"] = True
    config["database"]["uri"] = get_external_database_url(instance_type)

    return config


def add_database_config(config: dict, instance_type: InstanceType) -> dict:
    """
    Add the database configuration to the config
    """
    if instance_type == InstanceType.HQ:
        service_name = "HQ"
    elif instance_type == InstanceType.ALGORITHM_STORE:
        service_name = "store"
    elif instance_type == InstanceType.AUTH:
        service_name = "auth"
    else:
        raise ValueError(f"Invalid instance type: {instance_type}")

    # === Database settings ===
    config["database"]["volumePath"] = q.text(
        f"Where is your {service_name} database located on the host machine?",
        default=f"{Path.cwd()}/dev/.db/db_pv_{service_name}",
    ).unsafe_ask()

    return config


def get_external_database_url(instance_type: InstanceType) -> dict:
    """
    Add the production backend configuration to the config

    Parameters
    ----------
    config : dict
        The config to add the production backend configuration to
    instance_type : InstanceType
        Type of backend instance (HQ, algorithm store, or auth)

    Returns
    -------
    dict
        The config with the production backend configuration added
    """
    info("For production environments, it is recommended to use an external database.")
    info("Please provide the URI of the external database.")
    info("Example: postgresql://username:password@localhost:5432/vantage6")

    # Set different default URIs based on instance type
    if instance_type == InstanceType.ALGORITHM_STORE:
        default_uri = "postgresql://vantage6:vantage6@localhost:5433/vantage6_store"
    elif instance_type == InstanceType.AUTH:
        default_uri = "postgresql://vantage6:vantage6@localhost:5434/vantage6_auth"
    elif instance_type == InstanceType.HQ:
        default_uri = "postgresql://vantage6:vantage6@localhost:5432/vantage6"
    else:
        raise ValueError(
            f"Invalid instance type to set default database URI for: {instance_type}"
        )

    database_uri = q.text(
        "Database URI:",
        default=default_uri,
    ).unsafe_ask()

    parsed = urlparse(database_uri)
    if parsed.scheme not in ["postgresql", "postgres"]:
        error(f"Unsupported database scheme: {parsed.scheme}. Use postgresql://")
        exit(1)

    return database_uri


def make_configuration(
    config_producing_func: Callable,
    config_producing_func_args: tuple,
    type_: InstanceType,
    instance_name: str,
    system_folders: bool,
    is_sandbox: bool = False,
    extra_config: dict | None = None,
    save_config_file: bool = True,
    extra_configs_to_render: dict[str, str] | None = None,
) -> tuple[dict, Path]:
    """
    Create a configuration file for a vantage6 infrastructure component.

    Parameters
    ----------
    config_producing_func : Callable
        Function to generate the configuration
    config_producing_func_args : tuple
        Arguments to pass to the config producing function
    type_ : InstanceType
        Type of the instance to create a configuration for
    instance_name : str
        Name of the instance
    system_folders : bool
        Whether to use the system folders or not
    is_sandbox : bool
        Whether to create a sandbox configuration or not
    extra_config: dict | None = None
        Extra configuration to add. Note that this may overwrite the configuration
        produced by the config producing function if the keys overlap.
    save_config_file: bool = True,
        Whether to store the configuration on disk or not.
    extra_configs_to_render: dict[str, str] | None = None
        Extra configurations to render. These will be added to the configuration file.

    Returns
    -------
    tuple[dict, Path]
        Dictionary with the configuration and path to the configuration file
    """
    # for defaults and where to save the config
    dirs = AppContext.instance_folders(type_, instance_name, system_folders)

    # invoke function to create configuration file. Usually this is a questionaire
    # but it can also be a function that immediately returns a dict with the
    # configuration.
    config: dict = config_producing_func(*config_producing_func_args)

    if extra_config:
        config = merge_nested_dicts(config, extra_config)

    # in the case of an environment we need to add it to the current
    # configuration. In the case of application we can simply overwrite this
    # key (although there might be environments present)
    config_file = Path(dirs.get("config")) / (instance_name + ".yaml")
    if type_ == InstanceType.NODE:
        conf_manager = NodeConfigurationManager
    elif type_ == InstanceType.HQ:
        conf_manager = HQConfigurationManager
    elif type_ == InstanceType.ALGORITHM_STORE:
        conf_manager = AlgorithmStoreConfigurationManager
    elif type_ == InstanceType.AUTH:
        conf_manager = AuthConfigurationManager
    elif type_ == InstanceType.HUB:
        conf_manager = HubConfigurationManager
    else:
        raise ValueError(f"Invalid instance type: {type_}")

    if Path(config_file).exists():
        config_manager = conf_manager.from_file(config_file, is_sandbox=is_sandbox)
    else:
        config_manager = conf_manager(instance_name, is_sandbox=is_sandbox)

    config_manager.put(config)
    if save_config_file:
        config_file = config_manager.save(
            config_file, extra_configs_to_render=extra_configs_to_render
        )

    return config_manager.render_config(), config_file


def select_configuration_questionnaire(
    type_: InstanceType, system_folders: bool, is_sandbox: bool = False
) -> str:
    """
    Ask which configuration the user wants to use. It shows only configurations
    that are in the default folder.

    Parameters
    ----------
    type_ : InstanceType
        Type of the instance to create a configuration for
    system_folders : bool
        Whether to use the system folders or not
    is_sandbox : bool
        Whether to show only the sandbox configurations or not

    Returns
    -------
    str
        Name of the configuration
    """
    context = select_context_class(type_)
    configs, _ = context.available_configurations(system_folders, is_sandbox)

    # each collection (file) can contain multiple configs. (e.g. test,
    # dev)
    choices = []
    for config_collection in configs:
        choices.append(
            q.Choice(title=f"{config_collection.name:25}", value=config_collection.name)
        )

    if not choices:
        raise Exception("No configurations could be found!")

    # pop the question
    try:
        return q.select(
            "Select the configuration you want to use:", choices=choices
        ).unsafe_ask()
    except KeyboardInterrupt:
        error("Aborted by user!")
        exit(1)



import click
from kubernetes import client as k8s_client

from vantage6.common import error, info, warning
from vantage6.common.globals import (
    DEFAULT_NODE_IMAGE,
    DEFAULT_NODE_IMAGE_WO_TAG,
    InstanceType,
)

from vantage6.cli import __version__
from vantage6.cli.common.attach import attach_logs
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.start import (
    helm_install,
    prestart_checks,
)
from vantage6.cli.common.utils import create_directory_if_not_exists
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import ChartName, InfraComponentName
from vantage6.cli.k8s_config import KubernetesConfig, select_k8s_config
from vantage6.cli.node.common import create_client
from vantage6.cli.utils_kubernetes import get_core_api_with_ssl_handling


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--attach/--detach",
    default=False,
    help="Show node logs on the current console after starting the node",
)
@click.option("--local-chart-dir", default=None, help="Local chart directory to use")
@click.option("--chart-version", default=None, help="Chart version to use")
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(
    InstanceType.NODE,
    include_name=True,
    include_system_folders=True,
    sandbox_param="sandbox",
)
def cli_node_start(
    ctx: NodeContext,
    name: str,
    system_folders: bool,
    context: str,
    namespace: str,
    attach: bool,
    local_chart_dir: str,
    chart_version: str | None,
) -> None:
    """
    Start the node.
    """
    info("Starting node...")

    prestart_checks(ctx, InstanceType.NODE, name, system_folders)

    k8s_config = select_k8s_config(context=context, namespace=namespace)

    create_directory_if_not_exists(ctx.log_dir)
    create_directory_if_not_exists(ctx.data_dir)
    create_task_namespace_if_not_exists(ctx, k8s_config)

    # Determine image-name. First we check if the image is set in the config file.
    # If so, we use it. Otherwise, we use the same version as HQ.
    if custom_image := ctx.config.get("node", {}).get("image"):
        image = custom_image
    else:
        # if no custom image is specified, find HQ version and use
        # the latest images from that minor version
        client = create_client(ctx, use_sandbox_port=ctx.is_sandbox)
        major_minor = None
        try:
            # try to get HQ version, skip if can't get a connection
            version = client.util.get_hq_version(attempts_on_timeout=3)["version"]
            major_minor = ".".join(version.split(".")[:2])
            image = f"{DEFAULT_NODE_IMAGE_WO_TAG}:{major_minor}"
        except Exception:
            warning("Could not determine HQ version. Using default node image")

        if major_minor and not __version__.startswith(major_minor):
            warning(
                "Version mismatch between CLI and HQ/node. CLI is running on "
                f"version {__version__}, while node and HQ are on version "
                f"{major_minor}. This might cause unexpected issues; changing to "
                f"{major_minor}.<latest> is recommended."
            )

    # fail safe, in case no custom image is specified and we can't get the HQ version
    if not image:
        image = DEFAULT_NODE_IMAGE

    helm_install(
        release_name=ctx.helm_release_name,
        chart_name=ChartName.NODE,
        values_file=ctx.config_file,
        k8s_config=k8s_config,
        local_chart_dir=local_chart_dir,
        custom_values=[f"node.image={image}"],
        chart_version=chart_version,
    )

    if attach:
        attach_logs(
            name,
            instance_type=InstanceType.NODE,
            infra_component=InfraComponentName.NODE,
            system_folders=system_folders,
            k8s_config=k8s_config,
            is_sandbox=ctx.is_sandbox,
        )


def create_task_namespace_if_not_exists(
    ctx: NodeContext, k8s_config: KubernetesConfig
) -> None:
    """
    Create the task namespace if it does not exist.

    Parameters
    ----------
    ctx: NodeContext
        The context of the node.
    k8s_config: KubernetesConfig
        The Kubernetes configuration.
    """
    task_namespace = ctx.config.get("node", {}).get("taskNamespace")
    if not task_namespace:
        warning("Could not find node's task namespace in the config file.")
        exit(1)

    # TODO the namespace creation could use some refactoring, e.g. with v6 use namespace
    core_api = get_core_api_with_ssl_handling()
    try:
        namespace_list = core_api.list_namespace()
    except Exception:
        error(
            "Failed to list namespaces. Check if the cluster is running and reachable."
        )
        exit(1)
    namespace_names = [ns.metadata.name for ns in namespace_list.items]
    if task_namespace not in namespace_names:
        namespace_body = k8s_client.V1Namespace(
            metadata=k8s_client.V1ObjectMeta(name=task_namespace)
        )
        core_api.create_namespace(namespace_body)



import subprocess
import sys

import click

from vantage6.common import error, info

from vantage6.cli.dev.common import check_devspace_installed


@click.command()
@click.option("--only-hq", is_flag=True, help="Rebuild the HQ image.")
@click.option("--only-node", is_flag=True, help="Rebuild the node image.")
@click.option("--only-store", is_flag=True, help="Rebuild the store image.")
@click.option("--only-ui", is_flag=True, help="Rebuild the ui image.")
def cli_rebuild_dev_env(
    only_hq: bool, only_node: bool, only_store: bool, only_ui: bool
):
    """Rebuild Docker images for your development environment."""
    check_devspace_installed()

    try:
        info("🔄 Rebuilding development environment with devspace...")
        cmd = ["devspace", "run", "rebuild"]

        if only_hq:
            cmd.append("--hq")
        if only_node:
            cmd.append("--node")
        if only_store:
            cmd.append("--store")
        if only_ui:
            cmd.append("--ui")

        subprocess.run(cmd, check=True, capture_output=False)
        info("✅ Development environment rebuilt successfully!")
    except subprocess.CalledProcessError as e:
        error(f"❌ Error rebuilding development environment: {e}")
        sys.exit(e.returncode)



from pathlib import Path

import yaml

from vantage6.common.globals import InstanceType

from vantage6.cli.context import select_context_class
from vantage6.cli.sandbox.populate.helpers.utils import (
    replace_wsl_path,
    str_replace_wsl_path,
)


class BaseSandboxConfigManager:
    """
    Base class for sandbox configuration managers.

    Parameters
    ----------
    hq_name : str
        Name of the HQ.
    custom_data_dir : Path | None
        Path to the custom data directory. Useful on WSL because of mount issues for
        default directories.
    """

    def __init__(self, hub_name: str, custom_data_dir: Path | None) -> None:
        self.hub_name = hub_name
        self.custom_data_dir = Path(custom_data_dir) if custom_data_dir else None

    @staticmethod
    def _read_extra_config_file(extra_config_file: Path | None) -> dict:
        """Reads extra configuration file.

        Parameters
        ----------
        extra_config_file : Path | None
            Path to file with additional configuration.

        Returns
        -------
        dict
            Extra configuration parsed from YAML. Empty dict if none provided.
        """
        if extra_config_file:
            with open(extra_config_file, "r", encoding="utf-8") as f:
                loaded = yaml.safe_load(f) or {}
                if not isinstance(loaded, dict):
                    # Ensure we always return a dictionary
                    return {"value": loaded}
                return loaded
        return {}

    def _create_and_get_data_dir(
        self,
        instance_type: InstanceType,
        is_data_folder: bool = False,
        node_name: str | None = None,
        is_log_dir: bool = False,
        custom_folder: str | None = None,
    ) -> str:
        """
        Create and get the data directory.

        Parameters
        ----------
        instance_type: InstanceType
            Type of vantage6 component
        is_data_folder: bool
            Whether or not to create the data folder or a config folder. This is only
            used for node databases. Default is False.
        node_name: str | None
            Name of the node. Only used if is_data_folder is True.
        is_log_dir: bool
            Whether or not to create a log directory instead of a data directory.
            Default is False.
        custom_folder: str | None
            Name of the custom folder to create. Default is None.

        Returns
        -------
        str
            Path to the created directory
        """
        ctx_class = select_context_class(instance_type)
        folders = ctx_class.instance_folders(
            instance_type=InstanceType.HUB,
            instance_name=self.hub_name,
            system_folders=False,
        )
        if self.custom_data_dir:
            main_data_dir = self.custom_data_dir
        elif is_log_dir:
            main_data_dir = Path(folders["log"])
        else:
            main_data_dir = Path(folders["dev"])

        if instance_type == InstanceType.HQ:
            folder_name = custom_folder or "hq"
            data_dir = main_data_dir / self.hub_name / folder_name
        elif instance_type == InstanceType.ALGORITHM_STORE:
            folder_name = custom_folder or "store"
            data_dir = main_data_dir / self.hub_name / folder_name
        elif instance_type == InstanceType.NODE:
            if custom_folder:
                last_subfolder = custom_folder
            elif is_data_folder:
                if node_name:
                    last_subfolder = f"data_{node_name}"
                else:
                    last_subfolder = "data"
            else:
                last_subfolder = "node"
            data_dir = main_data_dir / self.hub_name / last_subfolder
        elif instance_type == InstanceType.AUTH:
            folder_name = custom_folder or "auth"
            data_dir = main_data_dir / self.hub_name / folder_name
        else:
            raise ValueError(f"Invalid instance type to get data dir: {instance_type}")

        # For the directory to be created, ensure that if a WSL path is used, the path
        # is converted to /mnt/wsl to create the directory on the host (not
        # /run/desktop/mnt/host/wsl as will raise non-existent directory errors)
        data_dir = replace_wsl_path(data_dir, to_mnt_wsl=True)
        data_dir.mkdir(parents=True, exist_ok=True)
        # now ensure that the wsl path is properly replaced to /run/desktop/mnt/host/wsl
        # if it is a WSL path, because that path will be used in the node configuration
        # files and is required to successfully mount the volumes.
        return str_replace_wsl_path(data_dir, to_mnt_wsl=False)



import click
from kubernetes.config.config_exception import ConfigException
from kubernetes.stream import stream

from vantage6.common import error, info
from vantage6.common.globals import APPNAME, InstanceType

from vantage6.cli import __version__
from vantage6.cli.common.utils import (
    extract_name_and_is_sandbox,
)
from vantage6.cli.common.version import get_and_select_ctx
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.k8s_config import KubernetesConfig, select_k8s_config
from vantage6.cli.utils_kubernetes import get_core_api_with_ssl_handling


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders rather than user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in user folders rather than "
    "system folders. This is the default",
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--sandbox", "is_sandbox", flag_value=True, help="Is this a sandbox environment?"
)
def cli_node_version(
    name: str, system_folders: bool, context: str, namespace: str, is_sandbox: bool
) -> None:
    """
    Returns current version of a vantage6 node.
    """
    name, is_sandbox = extract_name_and_is_sandbox(name, is_sandbox)
    if is_sandbox:
        system_folders = False

    k8s_config = select_k8s_config(context=context, namespace=namespace)

    ctx = get_and_select_ctx(
        InstanceType.NODE, name, system_folders, k8s_config, is_sandbox
    )
    version = _get_node_version_from_k8s(ctx.helm_release_name, k8s_config)
    info("")
    info(f"Node version: {version}")
    info(f"CLI version: {__version__}")


def _get_node_version_from_k8s(
    helm_release: str,
    k8s_config: KubernetesConfig,
) -> str:
    """
    Runs 'vnode-local version' in the node pod belonging to the Helm release.
    """
    pod = _get_pod_name_for_helm_release(helm_release, k8s_config)
    output = _exec_pod_command(
        pod_name=pod,
        namespace=k8s_config.namespace,
        command=["vnode-local", "version"],
    )
    return output.strip()


def _get_pod_name_for_helm_release(
    helm_release: str,
    k8s_config: KubernetesConfig,
) -> str:
    """
    Returns the first pod name for a given Helm release in a namespace.
    Looks up pods using the standard Helm label 'app.kubernetes.io/instance'.
    """
    try:
        # Load kubeconfig (context optional). Falls back to in-cluster if not available.
        try:
            k8s_config.load_kube_config(context=k8s_config.context)  # desktop/dev
        except ConfigException:
            k8s_config.load_incluster_config()  # in-cluster
    except ConfigException as exc:
        raise RuntimeError(f"Failed to load Kubernetes config: {exc}") from exc

    core = get_core_api_with_ssl_handling()
    selector = f"app={APPNAME}-node,release={helm_release}"
    pods = core.list_namespaced_pod(
        namespace=k8s_config.namespace, label_selector=selector
    ).items
    if not pods:
        error(
            f"No pods found for Helm release '{helm_release}' in ns "
            f"'{k8s_config.namespace}'"
        )
        exit(1)
    # Prefer a Ready pod
    for p in pods:
        for cond in p.status.conditions or []:
            if cond.type == "Ready" and cond.status == "True":
                return p.metadata.name
    # Fallback to first pod
    return pods[0].metadata.name


def _exec_pod_command(
    pod_name: str,
    namespace: str,
    command: list[str],
) -> str:
    """
    Executes a command inside the specified pod (and optional container) and returns stdout.
    """
    core = get_core_api_with_ssl_handling()
    resp = stream(
        core.connect_get_namespaced_pod_exec,
        pod_name,
        namespace,
        command=command,
        stderr=True,
        stdin=False,
        stdout=True,
        tty=False,
    )
    return resp



import subprocess

import click
import questionary as q

from vantage6.common import error, warning
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import (
    find_running_service_names,
    get_config_name_from_helm_release_name,
)
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.k8s_config import select_k8s_config
from vantage6.cli.node.stop import cli_node_stop


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders instead of user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in the user folders instead of "
    "system folders. This is the default.",
)
@click.option(
    "--attach/--detach",
    default=False,
    help="Show node logs on the current console after starting the node",
)
@click.option("--all", "all_nodes", flag_value=True, help="Stop all running nodes")
@click.option("--sandbox", "is_sandbox", flag_value=True, help="Restart a sandbox node")
@click.pass_context
def cli_node_restart(
    click_ctx: click.Context,
    name: str,
    context: str,
    namespace: str,
    system_folders: bool,
    attach: bool,
    all_nodes: bool,
    is_sandbox: bool,
) -> None:
    """Restart the node"""
    k8s_config = select_k8s_config(context=context, namespace=namespace)

    running_node_names = find_running_service_names(
        instance_type=InstanceType.NODE,
        only_system_folders=system_folders,
        only_user_folders=not system_folders,
        k8s_config=k8s_config,
    )
    if not running_node_names:
        warning("No nodes are currently running. No action taken.")
        return

    if attach and all_nodes:
        error(
            "Cannot attach logs of all nodes at once. Please remove either the "
            "'--all' or '--attach' option."
        )
        return

    if all_nodes:
        names = [
            get_config_name_from_helm_release_name(container_name)
            for container_name in running_node_names
        ]
    else:
        if not name:
            try:
                helm_name = q.select(
                    "Select the node you wish to restart:", choices=running_node_names
                ).unsafe_ask()
            except KeyboardInterrupt:
                error("Aborted by user!")
                return
            names = [get_config_name_from_helm_release_name(helm_name)]
        else:
            names = [name]

    for node_name in names:
        click_ctx.invoke(
            cli_node_stop,
            name=node_name,
            system_folders=system_folders,
            context=k8s_config.context,
            namespace=k8s_config.namespace,
            all_nodes=False,
            is_sandbox=is_sandbox,
        )

        cmd = [
            "v6",
            "node",
            "start",
            "--name",
            node_name,
            "--context",
            k8s_config.context,
            "--namespace",
            k8s_config.namespace,
        ]
        if system_folders:
            cmd.append("--system")
        if attach:
            cmd.append("--attach")
        if is_sandbox:
            cmd.append("--sandbox")
        subprocess.run(cmd, check=True)



"""Command line interface for the vantage6 infrastructure."""

import importlib.metadata

# note that here we cannot use __package__ because __package__ resolves to vantage6.cli
# whereas the PyPi package is called vantage6
__version__ = importlib.metadata.version("vantage6")



import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from vantage6.common.globals import FILE_BASED_DATABASE_TYPES, InstanceType, NodePolicy

from vantage6.cli.configuration_create import (
    make_configuration,
    select_configuration_questionnaire,
)
from vantage6.cli.hq.new import hq_configuration_questionaire
from vantage6.cli.node.new import node_configuration_questionaire

module_path = "vantage6.cli.configuration_create"


class WizardTest(unittest.TestCase):
    @staticmethod
    def prompts(*args, **kwargs):
        result = {}
        for arg in args[0]:
            name = arg["name"]
            if name == "default":  # default db path
                result[name] = "/some/path/db.sqlite"
            else:
                if "default" in arg:
                    result[name] = arg["default"]
                else:
                    result[name] = None
        return result

    @patch("vantage6.cli.node.new.q")
    @patch("vantage6.cli.node.new.NodeClient.authenticate")
    def test_node_wizard(self, authenticate, q):
        """Check the node wizard runs without errors."""
        q.unsafe_prompt.side_effect = self.prompts
        q.confirm.return_value.unsafe_ask.side_effect = [
            True,  # add a database
            "Database reachable by URI",  # type of database to add
            "http://localhost:5000",  # URI of database
            FILE_BASED_DATABASE_TYPES[0],  # type of database
            False,  # don't add environment variables
            False,  # don't add another database
            True,  # add algorithm policies
            True,  # add list of allowed_algorithms
            "some-image",  # algorithm image to whitelist
            False,  # don't add another algorithm image
            True,  # add algorithm stores to allowed_algorithm_stores
            "some-store",  # algorithm store to whitelist
            False,  # don't add another algorithm store
            False,  # answer question on combining policies on store level and
            # single algorithm level
            "DEBUG",  # level of logging
            True,  # Enable encryption
            "privkey_Test.pem",  # path to private key file
        ]
        authenticate.side_effect = None
        config = node_configuration_questionaire(
            data_dir="/data", instance_name="iknl", log_dir="/log"
        )

        keys = [
            "api_key",
            "api_path",
            "databases",
            "encryption",
            "logging",
            "port",
            "hq_url",
            "task_dir",
        ]
        for key in keys:
            self.assertIn(key, config["node"])
        nested_keys = [
            ["policies", NodePolicy.ALLOWED_ALGORITHMS.value],
            ["policies", NodePolicy.ALLOWED_ALGORITHM_STORES.value],
            ["encryption", "private_key"],
            ["encryption", "enabled"],
            ["logging", "loggers"],
            ["databases", "fileBased"],
            ["databases", "serviceBased"],
        ]
        for nesting in nested_keys:
            current_config = config["node"]
            for key in nesting:
                self.assertIn(key, current_config)
                current_config = current_config[key]
        assert len(config["node"]["databases"]["fileBased"]) == 0
        assert len(config["node"]["databases"]["serviceBased"]) == 1

    @patch("vantage6.cli.configuration_create.AppContext")
    @patch("vantage6.cli.configuration_create.q")
    @patch("vantage6.cli.hq.new.q")
    def test_hq_wizard(self, q, q_hq_common, app_context):
        app_context.instance_folders.return_value = {"log": "/log"}
        q.unsafe_prompt.side_effect = self.prompts
        q_hq_common.unsafe_prompt.side_effect = self.prompts
        q_hq_common.text.return_value.unsafe_ask.side_effect = [
            "postgresql://vantage6:vantage6@localhost:5432/vantage6",  # URI of database
        ]
        q.confirm.return_value.unsafe_ask.side_effect = [
            True,  # setup allowed algorithm stores
            "https://store.uluru.vantage6.ai",  # allowed algorithm stores
        ]

        config = hq_configuration_questionaire(
            instance_name="vtg6",
            system_folders=False,
        )

        keys = ["database", "rabbitmq", "hq", "ui"]
        for key in keys:
            self.assertIn(key, config)
        nested_keys = [
            ["database", "external"],
            ["database", "uri"],
            ["hq", "api_path"],
            ["hq", "keycloak"],
            ["hq", "logging"],
            ["ui", "allowedAlgorithmStores"],
            ["rabbitmq", "password"],
        ]
        for nesting in nested_keys:
            current_config = config
            for key in nesting:
                self.assertIn(key, current_config)
                current_config = current_config[key]

    @patch("vantage6.cli.node.new.node_configuration_questionaire")
    @patch("vantage6.cli.hq.new.hq_configuration_questionaire")
    @patch("vantage6.cli.configuration_create.HQConfigurationManager")
    @patch("vantage6.cli.configuration_create.NodeConfigurationManager")
    @patch("vantage6.cli.configuration_create.AppContext")
    def test_configuration_create_interface(self, context, node_m, hq_m, hq_q, node_q):
        context.instance_folders.return_value = {"config": "/some/path/"}
        # Configure mocks to return whatever path is passed as argument - this is
        # necessary as it converts the path to a Path object and that generates Mocking
        # issues otherwise
        node_m.return_value.save.side_effect = lambda path, **kwargs: path
        hq_m.return_value.save.side_effect = lambda path, **kwargs: path

        _, file_ = make_configuration(
            config_producing_func=node_q,
            config_producing_func_args=("/some/path/", "vtg6"),
            type_=InstanceType.NODE,
            instance_name="vtg6",
            system_folders=False,
        )
        self.assertEqual(Path("/some/path/vtg6.yaml"), file_)

        _, file_ = make_configuration(
            config_producing_func=hq_q,
            config_producing_func_args=("vtg6",),
            type_=InstanceType.HQ,
            instance_name="vtg6",
            system_folders=True,
        )
        self.assertEqual(Path("/some/path/vtg6.yaml"), file_)

    @patch("vantage6.cli.configuration_create.AppContext.available_configurations")
    def test_select_configuration(self, available_configurations):
        config = MagicMock()
        config.name = "vtg6"

        available_configurations.return_value = [[config], []]

        with patch(f"{module_path}.q") as q:
            q.select.return_value.unsafe_ask.return_value = "vtg6"
            name = select_configuration_questionnaire(InstanceType.NODE, True)

        self.assertEqual(name, "vtg6")



import time
from pathlib import Path

import click
from colorama import Fore, Style

from vantage6.common import error, info, warning
from vantage6.common.globals import HTTP_LOCALHOST, InstanceType, Ports

from vantage6.client import Client
from vantage6.client.utils import LogLevel

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.start import execute_cli_start
from vantage6.cli.context.hub import HubContext
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import CLICommandName
from vantage6.cli.k8s_config import KubernetesConfig, select_k8s_config
from vantage6.cli.sandbox.config.node import NodeDataset, NodeSandboxConfigManager
from vantage6.cli.sandbox.populate import populate_hub_sandbox


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--local-chart-dir",
    type=click.Path(exists=True),
    default=None,
    help="Local chart repository to use.",
)
@click.option(
    "--re-initialize",
    is_flag=True,
    default=False,
    help="Re-initialize the sandbox. This will repopulate HQ and create new "
    "node configurations.",
)
@click.option(
    "--num-nodes",
    type=int,
    default=3,
    help="Generate this number of nodes in the development network. Only used if "
    "--re-initialize flag is provided.",
)
@click.option(
    "--node-image",
    type=str,
    default=None,
    help="Node image to use. Only used if --re-initialize flag is provided.",
)
@click.option(
    "--extra-node-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional node configuration. This will be "
    "appended to each of the node configuration files. Only used if --re-initialize "
    "flag is provided",
)
@click.option(
    "--add-dataset",
    type=(str, click.Path()),
    default=None,
    multiple=True,
    help="Add a dataset to the nodes. The first argument is the label of the database, "
    "the second is the path to the dataset file. Only used if the --re-initialize flag "
    "is provided.",
)
@click.option(
    "--data-dir",
    "custom_data_dir",
    type=click.Path(exists=True),
    default=None,
    help="Path to a custom data directory to use. This option is especially useful "
    "on WSL because of mount issues for default directories. Only used if the "
    "--re-initialize flag is provided.",
)
@click_insert_context(type_=InstanceType.HUB, is_sandbox=True)
def cli_sandbox_start(
    ctx: HubContext,
    context: str | None,
    namespace: str | None,
    local_chart_dir: Path | None,
    re_initialize: bool,
    num_nodes: int,
    node_image: str | None,
    extra_node_config: Path | None,
    add_dataset: tuple[str, Path] | None,
    custom_data_dir: Path | None,
) -> None:
    """
    Start a sandbox environment.
    """
    k8s_config = select_k8s_config(context=context, namespace=namespace)

    # TODO if re-initalize is specified, we must remove the existing node configs
    execute_sandbox_start(
        ctx=ctx,
        hub_name=ctx.name,
        k8s_config=k8s_config,
        num_nodes=num_nodes,
        initialize=re_initialize,
        node_image=node_image,
        extra_node_config=extra_node_config,
        add_dataset=add_dataset,
        custom_data_dir=custom_data_dir,
        local_chart_dir=local_chart_dir,
    )


def execute_sandbox_start(
    ctx: HubContext,
    hub_name: str,
    k8s_config: KubernetesConfig,
    num_nodes: int,
    initialize: bool,
    node_image: str | None = None,
    extra_node_config: Path | None = None,
    add_dataset: tuple[str, Path] | None = None,
    custom_data_dir: Path | None = None,
    local_chart_dir: str | None = None,
) -> None:
    with_prometheus = (
        ctx.config.get("hq", {}).get("prometheus", {}).get("enabled", False)
    )

    execute_cli_start(
        command_name=CLICommandName.HUB,
        name=hub_name,
        k8s_config=k8s_config,
        local_chart_dir=local_chart_dir,
        system_folders=False,
        is_sandbox=True,
    )

    hq_url = f"{ctx.config['global']['urls']['external']['hq']}{ctx.config['hq']['hq']['apiPath']}"
    # TODO this wait may not be needed anymore, because the hub charts waits for the
    # hq to be ready almost
    _wait_for_hq_to_be_ready(hq_url)

    # Then we need to populate HQ
    if initialize:
        node_config_names = _initialize_sandbox(
            hq_url=hq_url,
            hub_name=hub_name,
            num_nodes=num_nodes,
            ctx=ctx,
            node_image=node_image,
            extra_node_config=extra_node_config,
            add_dataset=add_dataset,
            k8s_config=k8s_config,
            custom_data_dir=custom_data_dir,
            with_prometheus=with_prometheus,
        )
    else:
        node_configs, _ = NodeContext.available_configurations(
            system_folders=False, is_sandbox=True
        )
        node_config_names = [
            config.name
            for config in node_configs
            if config.name.startswith(f"{hub_name}-node-")
        ]

    # Then start the nodes
    info("Starting nodes")
    for node_config_name in node_config_names:
        execute_cli_start(
            command_name=CLICommandName.NODE,
            name=node_config_name,
            k8s_config=k8s_config,
            local_chart_dir=local_chart_dir,
            system_folders=False,
            is_sandbox=True,
        )

    # Print the authentication credentials
    _print_auth_credentials(hub_name)


def _initialize_sandbox(
    hq_url: str,
    hub_name: str,
    num_nodes: int,
    ctx: HubContext,
    node_image: str | None,
    extra_node_config: Path | None,
    add_dataset: tuple[str, Path] | None,
    k8s_config: KubernetesConfig,
    custom_data_dir: Path | None,
    with_prometheus: bool,
) -> list[str]:
    info("Populating vantage6 hub")
    node_details = populate_hub_sandbox(
        hq_url=hq_url,
        auth_url=f"{HTTP_LOCALHOST}:{Ports.SANDBOX_AUTH}",
        number_of_nodes=num_nodes,
    )

    api_keys = [node["api_key"] for node in node_details]
    node_names = [node["name"] for node in node_details]

    extra_dataset = (
        NodeDataset(
            label=add_dataset[0],
            path=add_dataset[1],
        )
        if add_dataset is not None and add_dataset != ()
        else None
    )

    # Create node config files from the nodes that were just registered in the HQ
    node_config_manager = NodeSandboxConfigManager(
        hub_name=hub_name,
        api_keys=api_keys,
        node_names=node_names,
        hq_port=ctx.config["hq"]["hq"]["port"],
        node_image=node_image,
        extra_node_config=extra_node_config,
        extra_dataset=extra_dataset,
        k8s_config=k8s_config,
        custom_data_dir=custom_data_dir,
        with_prometheus=with_prometheus,
    )
    node_config_manager.generate_node_configs()

    return node_config_manager.node_config_names


def _print_auth_credentials(hub_name: str) -> None:
    """
    Find user credentials to print, from the auth config file

    Parameters
    ----------
    hub_name : str
        Name of the hub.
    """
    hub_ctx = HubContext(
        instance_name=hub_name,
        system_folders=False,
        is_sandbox=True,
    )
    try:
        keycloak_config = hub_ctx.config["global"]["keycloak"]
        username = keycloak_config["adminUsername"]
        password = keycloak_config["adminPassword"]
        info("--------------------------------")
        info("Login with the following credentials:")
        info(
            f"Open the browser at: {Fore.GREEN}{HTTP_LOCALHOST}:"
            f"{Ports.SANDBOX_UI.value}{Style.RESET_ALL}"
        )
        info(f"Username: {Fore.GREEN}{username}{Style.RESET_ALL}")
        info(f"Password: {Fore.GREEN}{password}{Style.RESET_ALL}")
        info("--------------------------------")
    except KeyError:
        warning("No user credentials found in the auth config.")


def _wait_for_hq_to_be_ready(hq_url: str) -> None:
    """
    Wait for the HQ to be initialized.

    Parameters
    ----------
    hq_url : str
        URL of the hq.
    """
    info("Waiting for HQ to become ready...")
    client = Client(
        hq_url=hq_url,
        auth_url=f"{HTTP_LOCALHOST}:{Ports.SANDBOX_AUTH}",
        log_level=LogLevel.ERROR,
    )
    max_retries = 100
    wait_time = 3
    ready = False
    for _ in range(max_retries):
        try:
            result = client.util.get_hq_health(silent_on_connection_error=True)
            if result and result.get("api"):
                info("HQ is ready.")
                ready = True
                break
        except Exception:
            info("Waiting for HQ to be ready...")
            time.sleep(wait_time)

    if not ready:
        error("HQ did not become ready in time. Exiting...")
        exit(1)



from pathlib import Path

from vantage6.common.globals import HTTP_LOCALHOST, InstanceType, Ports

from vantage6.cli.auth.new import parse_database_uri_to_config
from vantage6.cli.common.new import new
from vantage6.cli.k8s_config import KubernetesConfig
from vantage6.cli.sandbox.config.base import BaseSandboxConfigManager
from vantage6.cli.utils_kubernetes import replace_localhost_for_k8s


class SandboxHubConfigManager(BaseSandboxConfigManager):
    """
    Class to store the sandbox configurations.

    Parameters
    ----------
    hub_name : str
        Name of the HQ.
    hq_image : str | None
        Image of the HQ.
    store_image : str | None
        Image of the algorithm store.
    ui_image : str | None
        Image of the UI.
    extra_hq_config : Path | None
        Path to the extra HQ configuration file.
    extra_store_config : Path | None
        Path to the extra algorithm store configuration file.
    k8s_config : KubernetesConfig
        Kubernetes configuration.
    custom_data_dir : Path | None
        Path to the custom data directory. Useful on WSL because of mount issues for
        default directories.
    """

    def __init__(
        self,
        hub_name: str,
        hq_image: str | None,
        store_image: str | None,
        ui_image: str | None,
        extra_hq_config: Path | None,
        extra_store_config: Path | None,
        extra_auth_config: Path | None,
        k8s_config: KubernetesConfig,
        with_prometheus: bool = False,
        custom_data_dir: Path | None = None,
        external_db_uris: dict[str, str] | None = None,
    ) -> None:
        super().__init__(hub_name, custom_data_dir)
        self.hq_image = hq_image
        self.store_image = store_image
        self.ui_image = ui_image
        self.extra_hq_config = extra_hq_config
        self.extra_store_config = extra_store_config
        self.extra_auth_config = extra_auth_config
        self.k8s_config = k8s_config
        self.with_prometheus = with_prometheus
        self.external_db_uris = external_db_uris

        self.auth_config = None
        self.hq_config = None
        self.store_config = None

    def generate_hub_configs(self) -> None:
        """Generates the local sandbox network."""

        self._create_auth_config()

        self._create_hq_config()

        self._create_algo_store_config()

        # Additionally create a single hub configuration that combines the
        # individual auth, HQ and store configurations. This hub values file
        # is used with the parent "hub" Helm chart to deploy all components.
        self._create_hub_config()

    def _create_hub_config(self) -> None:
        """
        Create a hub configuration file (YAML) that aggregates the auth, HQ
        and algorithm store configurations, so the sandbox can be deployed
        via the single hub Helm chart.
        """
        # External URLs as seen from the host
        urls = {
            "external": {
                "auth": f"{HTTP_LOCALHOST}:{Ports.SANDBOX_AUTH.value}",
                "store": f"{HTTP_LOCALHOST}:{Ports.SANDBOX_ALGO_STORE.value}",
                "ui": f"{HTTP_LOCALHOST}:{Ports.SANDBOX_UI.value}",
                "hq": f"{HTTP_LOCALHOST}:{Ports.SANDBOX_HQ.value}",
            },
            # Internal URLs as seen inside the cluster. These match the service
            # names and ports created by the hub Helm chart.
            "internal": {
                "auth": (
                    f"http://vantage6-{self.hub_name}-user-hub-kc-service."
                    f"{self.k8s_config.namespace}.svc.cluster.local:"
                    f"{Ports.DEV_AUTH.value}"
                ),
                "store": (
                    f"http://vantage6-{self.hub_name}-user-hub-store."
                    f"{self.k8s_config.namespace}.svc.cluster.local:"
                    f"{Ports.SANDBOX_ALGO_STORE.value}"
                ),
                "ui": (
                    f"http://vantage6-{self.hub_name}-user-hub-ui."
                    f"{self.k8s_config.namespace}.svc.cluster.local:"
                    f"{Ports.SANDBOX_UI.value}"
                ),
                "hq": (
                    f"http://vantage6-{self.hub_name}-user-hub-hq."
                    f"{self.k8s_config.namespace}.svc.cluster.local:"
                    f"{Ports.SANDBOX_HQ.value}"
                ),
            },
        }

        # Shared Keycloak defaults used by the hub chart. Subcharts can still
        # override these where needed.
        keycloak = {
            "url": urls["external"]["auth"],
        }

        hub_config = {
            "urls": urls,
            "keycloak": keycloak,
            # Pass through the component-specific values so the hub chart's
            # subcharts receive the same configuration as the standalone charts.
            "hq": self.hq_config,
            "auth": self.auth_config,
            "store": self.store_config,
        }

        hub_config = new(
            config_producing_func=lambda: hub_config,
            config_producing_func_args=(),
            name=self.hub_name,
            system_folders=False,
            type_=InstanceType.HUB,
            is_sandbox=True,
            extra_configs_to_render={
                "hq_config": self.hq_config,
                "auth_config": self.auth_config,
                "store_config": self.store_config,
            },
        )

    def __hq_config_return_func(
        self,
        extra_config: dict,
    ) -> dict:
        """
        Return a dict with HQ configuration values to be used in creating the
        config file.

        Parameters
        ----------
        extra_config : dict
            Extra configuration (parsed from YAML) to be added to the HQ
            configuration.

        Returns
        -------
        dict
            Dictionary with HQ configuration values.
        """
        data_dir = self._create_and_get_data_dir(instance_type=InstanceType.HQ)

        log_dir = self._create_and_get_data_dir(InstanceType.HQ, is_log_dir=True)

        prometheus_config = {
            "enabled": self.with_prometheus,
        }
        if self.with_prometheus:
            prometheus_dir = self._create_and_get_data_dir(
                InstanceType.HQ, custom_folder="prometheus"
            )
            prometheus_config.update(
                {
                    "exporter_port": Ports.SANDBOX_PROMETHEUS.value,
                    "storageSize": "2Gi",
                    "storageClass": "local-storage",
                    "volumeHostPath": prometheus_dir,
                }
            )

        store_service = f"vantage6-{self.hub_name}-user-hub-store"
        store_address = (
            f"http://{store_service}.{self.k8s_config.namespace}.svc.cluster.local"
            f":{Ports.SANDBOX_ALGO_STORE.value}"
        )
        config = {
            "hq": {
                "baseUrl": f"{HTTP_LOCALHOST}:{Ports.SANDBOX_HQ.value}",
                "port": Ports.SANDBOX_HQ.value,
                "internal": {
                    "port": Ports.SANDBOX_HQ.value,
                },
                "image": (
                    self.hq_image or "harbor2.vantage6.ai/infrastructure/hq:uluru"
                ),
                "algorithm_stores": [
                    {
                        "name": "Local store",
                        "url": store_address,
                        "api_path": "/store",
                    }
                ],
                "logging": {
                    "level": "DEBUG",
                    "volumeHostPath": log_dir,
                },
                "jwt": {
                    "secret": "development-constant-secret!",
                },
                "dev": {
                    "host_uri": (
                        "host.docker.internal"
                        if self.k8s_config.k8s_node == "docker-desktop"
                        else "172.17.0.1"
                    ),
                    "store_address": store_address,
                    "forward_ports": True,
                    "local_hub_port_to_expose": Ports.SANDBOX_HQ.value,
                    "local_ui_port_to_expose": Ports.SANDBOX_UI.value,
                },
            },
            "rabbitmq": {},
            "database": (
                {
                    "external": True,
                    "uri": replace_localhost_for_k8s(
                        self.external_db_uris["hq"], self.k8s_config.k8s_node
                    ),
                    "k8sNodeName": self.k8s_config.k8s_node,
                }
                if self.external_db_uris
                else {
                    "volumePath": data_dir,
                    "k8sNodeName": self.k8s_config.k8s_node,
                }
            ),
            "ui": {
                "port": Ports.SANDBOX_UI.value,
                "image": (
                    self.ui_image or "harbor2.vantage6.ai/infrastructure/ui:uluru"
                ),
            },
            "prometheus": prometheus_config,
        }

        if self.with_prometheus:
            config["hq"]["dev"]["local_prometheus_port_to_expose"] = (
                Ports.SANDBOX_PROMETHEUS.value
            )

        # merge the extra config with the HQ config
        if extra_config is not None:
            config.update(extra_config)

        return config

    def _create_hq_config(self) -> None:
        """Creates HQ configuration file (YAML)."""

        extra_config = self._read_extra_config_file(self.extra_hq_config)
        if self.ui_image is not None:
            ui_config = extra_config.get("ui", {}) if extra_config is not None else {}
            ui_config["image"] = self.ui_image
            extra_config["ui"] = ui_config

        # Create the HQ config file
        self.hq_config = new(
            config_producing_func=self.__hq_config_return_func,
            config_producing_func_args=(extra_config,),
            name=self.hub_name,
            system_folders=False,
            type_=InstanceType.HQ,
            is_sandbox=True,
            save_config_file=False,
        )

    def _create_algo_store_config(self) -> None:
        """Create algorithm store configuration file (YAML)."""

        extra_config = self._read_extra_config_file(self.extra_store_config)

        data_dir = self._create_and_get_data_dir(InstanceType.ALGORITHM_STORE)
        log_dir = self._create_and_get_data_dir(
            InstanceType.ALGORITHM_STORE, is_log_dir=True
        )

        self.store_config = new(
            config_producing_func=self.__algo_store_config_return_func,
            config_producing_func_args=(extra_config, data_dir, log_dir),
            name=f"{self.hub_name}-store",
            system_folders=False,
            type_=InstanceType.ALGORITHM_STORE,
            is_sandbox=True,
            save_config_file=False,
        )

    def __algo_store_config_return_func(
        self, extra_config: dict, data_dir: str, log_dir: str
    ) -> dict:
        """
        Return a dict with algorithm store configuration values to be used in creating
        the config file.

        Returns
        -------
        dict
            Dictionary with algorithm store configuration values.
        """
        config = {
            "store": {
                "internal": {
                    "port": Ports.SANDBOX_ALGO_STORE.value,
                },
                "port": Ports.SANDBOX_ALGO_STORE.value,
                "logging": {
                    "level": "DEBUG",
                    "volumeHostPath": log_dir,
                },
                "image": (
                    self.store_image
                    or "harbor2.vantage6.ai/infrastructure/algorithm-store:uluru"
                ),
                "policies": {
                    "allowLocalhost": True,
                    "assignReviewOwnAlgorithm": True,
                },
                "dev": {
                    "host_uri": (
                        "host.docker.internal"
                        if self.k8s_config.k8s_node == "docker-desktop"
                        else "172.17.0.1"
                    ),
                    "disable_review": True,
                    "review_own_algorithm": True,
                    "forward_ports": True,
                    "local_port_to_expose": Ports.SANDBOX_ALGO_STORE.value,
                },
            },
            "database": (
                {
                    "external": True,
                    "uri": replace_localhost_for_k8s(
                        self.external_db_uris["store"], self.k8s_config.k8s_node
                    ),
                    "k8sNodeName": self.k8s_config.k8s_node,
                }
                if self.external_db_uris
                else {
                    "volumePath": data_dir,
                    "k8sNodeName": self.k8s_config.k8s_node,
                }
            ),
        }

        # merge the extra config with the algorithm store config
        if extra_config is not None:
            config.update(extra_config)

        return config

    def _create_auth_config(self) -> None:
        """Create auth configuration file (YAML)."""
        extra_config = self._read_extra_config_file(self.extra_auth_config)
        self.auth_config = new(
            config_producing_func=self.__auth_config_return_func,
            config_producing_func_args=(extra_config,),
            name=f"{self.hub_name}-auth",
            system_folders=False,
            type_=InstanceType.AUTH,
            is_sandbox=True,
            save_config_file=False,
        )

    def __auth_config_return_func(self, extra_config: dict) -> dict:
        """
        Return a dict with auth configuration values to be used in creating the
        config file.
        """

        data_dir = self._create_and_get_data_dir(InstanceType.AUTH)

        # Parse auth database URI to extract host, database name, username, password
        if self.external_db_uris:
            auth_uri = self.external_db_uris["auth"]
            database_config = parse_database_uri_to_config(
                auth_uri, k8s_node=self.k8s_config.k8s_node
            )
            database_config["k8sNodeName"] = self.k8s_config.k8s_node
        else:
            database_config = {
                "volumePath": data_dir,
                "k8sNodeName": self.k8s_config.k8s_node,
            }

        config = {
            "keycloak": {
                "production": False,
                "passwordUpdateRequired": False,
                "publicClient": {
                    "redirectUris": [
                        f"{HTTP_LOCALHOST}:{Ports.SANDBOX_UI.value}",
                    ],
                },
                "dev": {
                    "forward_ports": True,
                    "local_auth_port_to_expose": Ports.SANDBOX_AUTH.value,
                },
            },
            "database": database_config,
        }

        # merge the extra config with the auth config
        if extra_config is not None:
            config.update(extra_config)

        return config



import click

from vantage6.common import info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.context.node import NodeContext


@click.command()
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(type_=InstanceType.NODE, sandbox_param="sandbox")
def cli_node_files(ctx: NodeContext) -> None:
    """
    Prints the location of important node files.

    If the specified configuration cannot be found, it exits. Otherwise
    it returns the absolute path to the output.
    """
    info(f"Configuration file = {ctx.config_file}")
    info(f"Log file           = {ctx.log_file}")
    info(f"Data folders       = {ctx.data_dir}")
    info("Database labels and files")
    for db in ctx.databases.get("fileBased", []):
        info(
            f" - {db['name']:15} = {db['volumePath']}/{db['originalName']} "
            f"(type: {db['type']})"
        )
    for db in ctx.databases.get("serviceBased", []):
        info(f" - {db['name']:15} = {db['uri']} (type: {db['type']})")



from __future__ import annotations

from vantage6.common.context import AppContext
from vantage6.common.globals import InstanceType

from vantage6.cli.globals import (
    DEFAULT_API_SERVICE_SYSTEM_FOLDERS as S_FOL,
)


class HubContext(AppContext):
    """
    Hub context
    """

    def __init__(
        self, instance_name: str, system_folders: bool = S_FOL, is_sandbox: bool = False
    ):
        super().__init__(
            InstanceType.HUB,
            instance_name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
        )

    @classmethod
    def from_external_config_file(
        cls, path: str, system_folders: bool = S_FOL, in_container: bool = False
    ) -> HubContext:
        """
        Create a Hub context from an external configuration file.

        Parameters
        ----------
        path : str
            Path of the configuration file
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        in_container : bool, optional
            Whether the application is running inside a container, by default False

        Returns
        -------
        HubContext
            Hub context object
        """
        return super().from_external_config_file(
            path, InstanceType.HUB, system_folders, in_container
        )

    @classmethod
    def config_exists(
        cls, instance_name: str, system_folders: bool = S_FOL, is_sandbox: bool = False
    ) -> bool:
        """
        Check if a configuration file exists.

        Parameters
        ----------
        instance_name : str
            Name of the configuration instance, corresponds to the filename
            of the configuration file.
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        is_sandbox : bool, optional
            Whether the configuration is a sandbox configuration, by default False

        Returns
        -------
        bool
            Whether the configuration file exists or not
        """
        return super().base_config_exists(
            InstanceType.HUB,
            instance_name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
        )

    @classmethod
    def available_configurations(
        cls, system_folders: bool = S_FOL, is_sandbox: bool = False
    ) -> tuple[list, list]:
        """
        Find all available Hub configurations in the default folders.

        Parameters
        ----------
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        is_sandbox : bool, optional
            Whether the configuration is a sandbox configuration, by default False

        Returns
        -------
        tuple[list, list]
            The first list contains validated configuration files, the second
            list contains invalid configuration files.
        """
        return super().available_configurations(
            InstanceType.HUB, system_folders, is_sandbox
        )



import click

from vantage6.common import info, warning
from vantage6.common.globals import InstanceType

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.remove import execute_remove
from vantage6.cli.context.auth import AuthContext
from vantage6.cli.globals import InfraComponentName
from vantage6.cli.utils_kubernetes import get_core_api_with_ssl_handling


@click.command()
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(
    type_=InstanceType.AUTH,
    include_name=True,
    include_system_folders=True,
    sandbox_param="sandbox",
)
@click.option("-f", "--force", "force", flag_value=True)
def cli_auth_remove(
    ctx: AuthContext, name: str, system_folders: bool, force: bool
) -> None:
    """
    Function to remove an authentication service's data.

    Parameters
    ----------
    ctx : AuthContext
        Auth context object
    name : str
        Name of the auth service
    system_folders : bool
        Whether to use system folders or user folders
    force : bool
        Whether to ask for confirmation before removing or not
    """
    auth_remove(ctx, name, system_folders, force)


# this new function is just here so that it can be called from the sandbox remove
# command
def auth_remove(ctx: AuthContext, name: str, system_folders: bool, force: bool) -> None:
    # Best-effort cleanup of Keycloak PVCs and their bound PVs after uninstall
    try:
        _cleanup_auth_volumes(ctx)
    except Exception as e:
        # Cleanup is best-effort; do not fail the remove command if cleanup fails
        warning(f"Failed to cleanup auth volumes: {e}")

    execute_remove(
        ctx, InstanceType.AUTH, InfraComponentName.AUTH, name, system_folders, force
    )


def _cleanup_auth_volumes(ctx: AuthContext) -> None:
    core_api = get_core_api_with_ssl_handling()

    # Label used by the auth chart to tag its resources
    label_selector = f"app.kubernetes.io/instance={ctx.helm_release_name}"

    # Collect PVCs across all namespaces that belong to this release
    pvcs = core_api.list_persistent_volume_claim_for_all_namespaces(
        label_selector=label_selector
    ).items

    # Track PV names bound to these PVCs
    pv_names = {
        pvc.spec.volume_name for pvc in pvcs if pvc.spec and pvc.spec.volume_name
    }

    # Delete PVCs first (namespaced)
    for pvc in pvcs:
        ns = pvc.metadata.namespace
        name_ = pvc.metadata.name
        try:
            info(f"Deleting persistent volume claim {name_} in namespace {ns}")
            core_api.delete_namespaced_persistent_volume_claim(name=name_, namespace=ns)
        except Exception as e:
            # Ignore failures; continue attempting other deletions
            warning(
                f"Failed to delete persistent volume claim {name_} in namespace {ns}: "
                f"{e}"
            )

    # Delete PVs that were bound to those PVCs (cluster-scoped)
    for pv_name in pv_names:
        try:
            info(f"Deleting persistent volume: {pv_name}")
            core_api.delete_persistent_volume(name=pv_name)
        except Exception as e:
            warning(f"Failed to delete persistent volume {pv_name}: {e}")



import click
import questionary
from kubernetes import client

from vantage6.common import error

from vantage6.cli.utils import switch_context_and_namespace
from vantage6.cli.utils_kubernetes import (
    get_core_api_with_ssl_handling,
)


@click.command()
@click.argument("namespace", required=False, type=str, default=None)
def cli_use_namespace(namespace: str):
    """
    Set which Kubernetes namespace to use.

    The namespace will be created if it does not exist.
    """
    # Configure for MicroK8s if needed
    core_api = get_core_api_with_ssl_handling()

    try:
        namespace_list = core_api.list_namespace()
    except Exception:
        error(
            "Failed to connect to Kubernetes cluster. Check if the cluster is running "
            "and reachable."
        )
        return

    namespace_names = [ns.metadata.name for ns in namespace_list.items]

    # Prompt user to select a namespace
    if not namespace:
        namespace = questionary.select(
            "Which Kubernetes namespace do you want to use?",
            choices=namespace_names,
        ).ask()

        # If no namespace is selected (e.g. KeyboardInterrupt), exit
        if not namespace:
            return

    # Check if namespace exists, create if not
    if namespace not in namespace_names:
        namespace_body = client.V1Namespace(
            metadata=client.V1ObjectMeta(name=namespace)
        )
        core_api.create_namespace(namespace_body)

    # Switch to the selected namespace for current context
    switch_context_and_namespace(namespace=namespace)
